"""
Defines the [United States customary units][1] and their conversions to SI

[1]: https://en.wikipedia.org/wiki/United_States_customary_units


Attributes: Human-scale units of length

    Point (Unit): 1/72 inch
    Pica (Unit): 1/6 inch
    Inch (Unit): 1/12 foot, 25.4 millimeter
    Foot (Unit): 12 inches
    Yard (Unit): 3 feet
    Mile (Unit): 5,280 feet

Attributes: Survey units of length

    Link (Unit): 33/50 feet
    SurveyFoot (Unit): 1200/3937 meters
    Rod (Unit): 25 links
    Chain (Unit): 4 rods
    Furlong (Unit): 10 chains
    StatuteMile (Unit): 8 furlongs
    League (Unit): 3 statute miles

Attributes: Survey units of area and volume

    Acre (Unit): 43,560 ft²
    Section (Unit): 640 acres, 1 square statute mile
    SurveyTownship (Unit): 36 sections, 4 square leagues
    AcreFoot (Unit): a volume of 1 acre, 1 foot deep

Attributes: Nautical units of length

    Fathom (Unit): 6 feet
    Cable (Unit): 120 fathoms
    NauticalMile (Unit): 1,852 meters

"""

from . import (
    Acceleration,
    Area,
    Force,
    Length,
    Mass,
    Temperature,
    Unit,
    Volume,
    avoirdupois,
)
from .physics import gₙ
from .si import (
    Gram,
    Hour,
    Kelvin,
    Kilo,
    Liter,
    Meter,
    Micro,
    Milli,
    Newton,
    Pascal,
    Second,
)

# Length
# https://en.wikipedia.org/wiki/United_States_customary_units#Length

Point = Length.unit(name="point", symbol="p.")
Point.equals(127 / 360 * Milli * Meter)

Pica = Length.unit(name="pica", symbol="P.")
Pica.equals(12 * Point)
Pica.equals(127 / 30 * Milli * Meter)

Inch = Length.unit(name="inch", symbol="in.")
Inch.equals(6 * Pica)
Inch.equals(25.4 * Milli * Meter)

# https://en.wikipedia.org/wiki/Hand_(unit)
Hand = Length.unit(name="hand", symbol="hh")
Hand.equals(4 * Inch)

Foot = Length.unit(name="foot", symbol="ft.")
Foot.equals(12 * Inch)
Foot.equals(0.3048 * Meter)

Yard = Length.unit(name="yard", symbol="yd.")
Yard.equals(3 * Foot)
Yard.equals(36 * Inch)
Yard.equals(0.9144 * Meter)

Mile = Length.unit(name="mile", symbol="mi.")
Mile.equals(5280 * Foot)
Mile.equals(1760 * Yard)
Mile.equals(1.609344 * Kilo * Meter)

# US Survey Units

Link = Length.unit(name="link", symbol="li.")
Link.equals(7.92 * Inch)
Link.equals(33 / 50 * Foot)
Link.equals(792 / 3937 * Meter)

SurveyFoot = Length.unit(name="US survey foot", symbol="US-survey-foot")
SurveyFoot.equals(1200 / 3937 * Meter)
SurveyFoot.equals(1.000002 * Foot)

Rod = Length.unit(name="rod", symbol="rd.")
Rod.equals(25 * Link)
Rod.equals(16.5 * SurveyFoot)
Rod.equals(19800 / 3937 * Meter)

Chain = Length.unit(name="chain", symbol="ch.")
Chain.equals(4 * Rod)
Chain.equals(66 * SurveyFoot)
Chain.equals(79200 / 3937 * Meter)

Furlong = Length.unit(name="furlong", symbol="fur.")
Furlong.equals(10 * Chain)
Furlong.equals(792 / 3937 * Kilo * Meter)

StatuteMile = Length.unit(name="US statute mile", symbol="US-statute-mile")
StatuteMile.equals(8 * Furlong)
StatuteMile.equals(6336 / 3937 * Kilo * Meter)

League = Length.unit(name="league", symbol="lea.")
League.equals(3 * StatuteMile)
League.equals(19008 / 3937 * Kilo * Meter)

# Nautical Units

Fathom = Length.unit(name="fathom", symbol="ftm.")
Fathom.equals(2 * Yard)
Fathom.equals(6 * Foot)
Fathom.equals(1143 / 625 * Meter)

Cable = Length.unit(name="cable", symbol="cb.")
Cable.equals(120 * Fathom)
Cable.equals(3429 / 15625 * Kilo * Meter)

NauticalMile = Length.unit(name="nautical mile", symbol="nmi.")
NauticalMile.equals(1852 * Meter)

Knot = Unit.derive(NauticalMile / Hour, name="knot", symbol="kn")

# Area
# https://en.wikipedia.org/wiki/United_States_customary_units#Area

Acre = Area.unit(name="acre", symbol="acre")
Acre.equals(43560 * Foot**2)
Acre.equals(10 * Chain**2)

Section = Area.unit(name="section", symbol="section")
Section.equals(640 * Acre)
Section.equals(1 * StatuteMile**2)

SurveyTownship = Area.unit(name="survey township", symbol="twp.")
SurveyTownship.equals(36 * Section)
SurveyTownship.equals(4 * League**2)


# Volume
# https://en.wikipedia.org/wiki/United_States_customary_units#Volume

BoardFoot = Volume.unit(name="boardfoot", symbol="FBM")
BoardFoot.equals(1 * Foot * 1 * Foot * 1 * Inch)

Rick = Volume.unit(name="rick", symbol="rick")
Rick.equals(4 * Foot * 8 * Foot * 16 * Inch)

Cord = Volume.unit(name="cord", symbol="cord")
Cord.equals(4 * Foot * 8 * Foot * 4 * Foot)

AcreFoot = Volume.unit(name="acre-foot", symbol="acre-ft.")
AcreFoot.equals(43560 * Foot**3)

# Fluid Volume
# https://en.wikipedia.org/wiki/United_States_customary_units#Fluid_volume

Minim = Volume.unit(name="minim", symbol="min.")
Minim.equals(61.611519921875 * Micro * Liter)

FluidDram = Volume.unit(name="fluid dram", symbol="fl.dr.")
FluidDram.equals(60 * Minim)
FluidDram.equals(3.6966911953125 * Milli * Liter)

Teaspoon = Volume.unit(name="teaspoon", symbol="tsp.")
Teaspoon.equals(80 * Minim)
Teaspoon.equals(4.92892159375 * Milli * Liter)

Tablespoon = Volume.unit(name="tablespoon", symbol="tbsp.")
Tablespoon.equals(3 * Teaspoon)
Tablespoon.equals(4 * FluidDram)
Tablespoon.equals(14.78676478125 * Milli * Liter)

FluidOunce = Volume.unit(name="fluid ounce", symbol="fl.oz.")
FluidOunce.equals(2 * Tablespoon)
FluidOunce.equals(29.5735295625 * Milli * Liter)

Shot = Volume.unit(name="shot", symbol="jig.")
Shot.equals(1.5 * FluidOunce)
Shot.equals(3 * Tablespoon)
Shot.equals(44.36029434375 * Milli * Liter)

Gill = Volume.unit(name="gill", symbol="gi.")
Gill.equals(8 / 3 * Shot)
Gill.equals(4 * FluidOunce)
Gill.equals(118.29411825 * Milli * Liter)

Cup = Volume.unit(name="cup", symbol="c.")
Cup.equals(2 * Gill)
Cup.equals(8 * FluidOunce)
Cup.equals(236.5882365 * Milli * Liter)

Pint = Volume.unit(name="pint", symbol="pt.")
Pint.equals(2 * Cup)
Pint.equals(473.176473 * Milli * Liter)

Quart = Volume.unit(name="quart", symbol="qt.")
Quart.equals(2 * Pint)
Quart.equals(0.946352946 * Liter)

Pottle = Volume.unit(name="pottle", symbol="pot.")
Pottle.equals(2 * Quart)
Pottle.equals(1.892705892 * Liter)

Gallon = Volume.unit(name="gallon", symbol="gal.")
Gallon.equals(4 * Quart)
Gallon.equals(231 * Inch**3)
Gallon.equals(3.785411784 * Liter)

Barrel = Volume.unit(name="barrel", symbol="bbl.")
Barrel.equals(42 * Gallon)
Barrel.equals(158.987294928 * Liter)

Hogshead = Volume.unit(name="hogshead", symbol="hogshead")
Hogshead.equals(1.5 * Barrel)
Hogshead.equals(63 * Gallon)
Hogshead.equals(8.421875 * Foot**3)
Hogshead.equals(238.480942392 * Liter)

# Dry Volumes
# https://en.wikipedia.org/wiki/United_States_customary_units#Dry_volume

DryPint = Volume.unit(name="dry pint", symbol="dry-pint")
DryPint.equals(33.6003125 * Inch**3)
DryPint.equals(0.5506104713575 * Liter)

DryQuart = Volume.unit(name="dry quart", symbol="dry-quart")
DryQuart.equals(2 * DryPint)

DryGallon = Volume.unit(name="dry gallon", symbol="dry-gallon")
DryGallon.equals(2 * DryQuart)

Peck = Volume.unit(name="peck", symbol="pk.")
Peck.equals(2 * DryGallon)

Bushel = Volume.unit(name="bushel", symbol="bu.")
Bushel.equals(4 * Peck)
Bushel.equals(35.23907016688 * Liter)

DryBarrel = Volume.unit(name="dry barrel", symbol="dry-barrel")
DryBarrel.equals(7056 * Inch**3)


# Mass
# https://en.wikipedia.org/wiki/United_States_customary_units#Mass_and_weight

# The US uses the avoirdupois system of masses, with some adjustments
Dram = avoirdupois.Dram
Grain = avoirdupois.Grain
Ounce = avoirdupois.Ounce
Pound = avoirdupois.Pound

Hundredweight = Mass.unit("US hundredweight", "cwt.")
Hundredweight.equals(100 * Pound)
Hundredweight.equals(45.359237 * Kilo * Gram)

Ton = Mass.unit("short ton", "ton")
Ton.equals(20 * Hundredweight)
Ton.equals(907.18474 * Kilo * Gram)

GForce = Acceleration.unit("g-force", "g-force")
GForce.equals(gₙ)
GForce.equals(32.17405 * Foot / Second**2)

PoundForce = Force.unit("pound-force", "lbf")
PoundForce.equals(1 * Pound * gₙ)
PoundForce.equals(1 * Pound * GForce)
PoundForce.equals(4.4482216152605 * Newton)

PSI = Unit.derive(PoundForce / Inch**2, "pounds per square inch", "psi")
PSI.equals(6894.757 * Pascal)

# Temperature
# https://en.wikipedia.org/wiki/Rankine_scale
# https://en.wikipedia.org/wiki/Fahrenheit

Rankine = Temperature.unit("Rankine", "R")
Rankine.equals(5 / 9 * Kelvin)

Fahrenheit = Temperature.scale(+459.67 * Rankine, name="fahrenheit", symbol="°F")
Fahrenheit.alias(symbol="degF")



# The file was automatically generated by Lark v1.1.2
__version__ = "1.1.2"

#
#
#   Lark Stand-alone Generator Tool
# ----------------------------------
# Generates a stand-alone LALR(1) parser
#
# Git:    https://github.com/erezsh/lark
# Author: Erez Shinan (erezshin@gmail.com)
#
#
#    >>> LICENSE
#
#    This tool and its generated code use a separate license from Lark,
#    and are subject to the terms of the Mozilla Public License, v. 2.0.
#    If a copy of the MPL was not distributed with this
#    file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#    If you wish to purchase a commercial license for this tool and its
#    generated code, you may contact me via email or otherwise.
#
#    If MPL2 is incompatible with your free or open-source project,
#    contact me and we'll work it out.
#
#

from abc import ABC, abstractmethod
from collections.abc import Sequence
from types import ModuleType
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Collection,
    Dict,
    FrozenSet,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
)
from typing import Pattern as REPattern
from typing import Set, Tuple, Type, TypeVar, Union


class LarkError(Exception):
    pass


class ConfigurationError(LarkError, ValueError):
    pass


def assert_config(value, options: Collection, msg="Got %r, expected one of %s"):
    if value not in options:
        raise ConfigurationError(msg % (value, options))


class GrammarError(LarkError):
    pass


class ParseError(LarkError):
    pass


class LexError(LarkError):
    pass


T = TypeVar("T")


class UnexpectedInput(LarkError):
    # --
    line: int
    column: int
    pos_in_stream = None
    state: Any
    _terminals_by_name = None

    def get_context(self, text: str, span: int = 40) -> str:
        # --
        assert self.pos_in_stream is not None, self
        pos = self.pos_in_stream
        start = max(pos - span, 0)
        end = pos + span
        if not isinstance(text, bytes):
            before = text[start:pos].rsplit("\n", 1)[-1]
            after = text[pos:end].split("\n", 1)[0]
            return before + after + "\n" + " " * len(before.expandtabs()) + "^\n"
        else:
            before = text[start:pos].rsplit(b"\n", 1)[-1]
            after = text[pos:end].split(b"\n", 1)[0]
            return (
                before + after + b"\n" + b" " * len(before.expandtabs()) + b"^\n"
            ).decode("ascii", "backslashreplace")

    def match_examples(
        self,
        parse_fn: "Callable[[str], Tree]",
        examples: Union[Mapping[T, Iterable[str]], Iterable[Tuple[T, Iterable[str]]]],
        token_type_match_fallback: bool = False,
        use_accepts: bool = True,
    ) -> Optional[T]:
        # --
        assert self.state is not None, "Not supported for this exception"

        if isinstance(examples, Mapping):
            examples = examples.items()

        candidate = (None, False)
        for i, (label, example) in enumerate(examples):
            assert not isinstance(example, str), "Expecting a list"

            for j, malformed in enumerate(example):
                try:
                    parse_fn(malformed)
                except UnexpectedInput as ut:
                    if ut.state == self.state:
                        if (
                            use_accepts
                            and isinstance(self, UnexpectedToken)
                            and isinstance(ut, UnexpectedToken)
                            and ut.accepts != self.accepts
                        ):
                            logger.debug(
                                "Different accepts with same state[%d]: %s != %s at example [%s][%s]"
                                % (self.state, self.accepts, ut.accepts, i, j)
                            )
                            continue
                        if isinstance(
                            self, (UnexpectedToken, UnexpectedEOF)
                        ) and isinstance(ut, (UnexpectedToken, UnexpectedEOF)):
                            if ut.token == self.token:  ##

                                logger.debug("Exact Match at example [%s][%s]" % (i, j))
                                return label

                            if token_type_match_fallback:
                                ##

                                if (ut.token.type == self.token.type) and not candidate[
                                    -1
                                ]:
                                    logger.debug(
                                        "Token Type Fallback at example [%s][%s]"
                                        % (i, j)
                                    )
                                    candidate = label, True

                        if candidate[0] is None:
                            logger.debug(
                                "Same State match at example [%s][%s]" % (i, j)
                            )
                            candidate = label, False

        return candidate[0]

    def _format_expected(self, expected):
        if self._terminals_by_name:
            d = self._terminals_by_name
            expected = [
                d[t_name].user_repr() if t_name in d else t_name for t_name in expected
            ]
        return "Expected one of: \n\t* %s\n" % "\n\t* ".join(expected)


class UnexpectedEOF(ParseError, UnexpectedInput):
    # --
    expected: "List[Token]"

    def __init__(self, expected, state=None, terminals_by_name=None):
        super(UnexpectedEOF, self).__init__()

        self.expected = expected
        self.state = state
        from .lexer import Token

        self.token = Token("<EOF>", "")  ##

        self.pos_in_stream = -1
        self.line = -1
        self.column = -1
        self._terminals_by_name = terminals_by_name

    def __str__(self):
        message = "Unexpected end-of-input. "
        message += self._format_expected(self.expected)
        return message


class UnexpectedCharacters(LexError, UnexpectedInput):
    # --

    allowed: Set[str]
    considered_tokens: Set[Any]

    def __init__(
        self,
        seq,
        lex_pos,
        line,
        column,
        allowed=None,
        considered_tokens=None,
        state=None,
        token_history=None,
        terminals_by_name=None,
        considered_rules=None,
    ):
        super(UnexpectedCharacters, self).__init__()

        ##

        self.line = line
        self.column = column
        self.pos_in_stream = lex_pos
        self.state = state
        self._terminals_by_name = terminals_by_name

        self.allowed = allowed
        self.considered_tokens = considered_tokens
        self.considered_rules = considered_rules
        self.token_history = token_history

        if isinstance(seq, bytes):
            self.char = seq[lex_pos : lex_pos + 1].decode("ascii", "backslashreplace")
        else:
            self.char = seq[lex_pos]
        self._context = self.get_context(seq)

    def __str__(self):
        message = (
            "No terminal matches '%s' in the current parser context, at line %d col %d"
            % (self.char, self.line, self.column)
        )
        message += "\n\n" + self._context
        if self.allowed:
            message += self._format_expected(self.allowed)
        if self.token_history:
            message += "\nPrevious tokens: %s\n" % ", ".join(
                repr(t) for t in self.token_history
            )
        return message


class UnexpectedToken(ParseError, UnexpectedInput):
    # --

    expected: Set[str]
    considered_rules: Set[str]
    interactive_parser: "InteractiveParser"

    def __init__(
        self,
        token,
        expected,
        considered_rules=None,
        state=None,
        interactive_parser=None,
        terminals_by_name=None,
        token_history=None,
    ):
        super(UnexpectedToken, self).__init__()

        ##

        self.line = getattr(token, "line", "?")
        self.column = getattr(token, "column", "?")
        self.pos_in_stream = getattr(token, "start_pos", None)
        self.state = state

        self.token = token
        self.expected = expected  ##

        self._accepts = NO_VALUE
        self.considered_rules = considered_rules
        self.interactive_parser = interactive_parser
        self._terminals_by_name = terminals_by_name
        self.token_history = token_history

    @property
    def accepts(self) -> Set[str]:
        if self._accepts is NO_VALUE:
            self._accepts = (
                self.interactive_parser and self.interactive_parser.accepts()
            )
        return self._accepts

    def __str__(self):
        message = "Unexpected token %r at line %s, column %s.\n%s" % (
            self.token,
            self.line,
            self.column,
            self._format_expected(self.accepts or self.expected),
        )
        if self.token_history:
            message += "Previous tokens: %r\n" % self.token_history

        return message


class VisitError(LarkError):
    # --

    obj: "Union[Tree, Token]"
    orig_exc: Exception

    def __init__(self, rule, obj, orig_exc):
        message = 'Error trying to process rule "%s":\n\n%s' % (rule, orig_exc)
        super(VisitError, self).__init__(message)

        self.rule = rule
        self.obj = obj
        self.orig_exc = orig_exc


class MissingVariableError(LarkError):
    pass


import logging
import re
import sys

logger: logging.Logger = logging.getLogger("lark")
logger.addHandler(logging.StreamHandler())
##

##

logger.setLevel(logging.CRITICAL)


NO_VALUE = object()


def classify(seq, key=None, value=None):
    d = {}
    for item in seq:
        k = key(item) if (key is not None) else item
        v = value(item) if (value is not None) else item
        if k in d:
            d[k].append(v)
        else:
            d[k] = [v]
    return d


def _deserialize(data, namespace, memo):
    if isinstance(data, dict):
        if "__type__" in data:  ##

            class_ = namespace[data["__type__"]]
            return class_.deserialize(data, memo)
        elif "@" in data:
            return memo[data["@"]]
        return {
            key: _deserialize(value, namespace, memo) for key, value in data.items()
        }
    elif isinstance(data, list):
        return [_deserialize(value, namespace, memo) for value in data]
    return data


class Serialize:
    # --

    def memo_serialize(self, types_to_memoize):
        memo = SerializeMemoizer(types_to_memoize)
        return self.serialize(memo), memo.serialize()

    def serialize(self, memo=None):
        if memo and memo.in_types(self):
            return {"@": memo.memoized.get(self)}

        fields = getattr(self, "__serialize_fields__")
        res = {f: _serialize(getattr(self, f), memo) for f in fields}
        res["__type__"] = type(self).__name__
        if hasattr(self, "_serialize"):
            self._serialize(res, memo)
        return res

    @classmethod
    def deserialize(cls, data, memo):
        namespace = getattr(cls, "__serialize_namespace__", [])
        namespace = {c.__name__: c for c in namespace}

        fields = getattr(cls, "__serialize_fields__")

        if "@" in data:
            return memo[data["@"]]

        inst = cls.__new__(cls)
        for f in fields:
            try:
                setattr(inst, f, _deserialize(data[f], namespace, memo))
            except KeyError as e:
                raise KeyError("Cannot find key for class", cls, e)

        if hasattr(inst, "_deserialize"):
            inst._deserialize()

        return inst


class SerializeMemoizer(Serialize):
    # --

    __serialize_fields__ = ("memoized",)

    def __init__(self, types_to_memoize):
        self.types_to_memoize = tuple(types_to_memoize)
        self.memoized = Enumerator()

    def in_types(self, value):
        return isinstance(value, self.types_to_memoize)

    def serialize(self):
        return _serialize(self.memoized.reversed(), None)

    @classmethod
    def deserialize(cls, data, namespace, memo):
        return _deserialize(data, namespace, memo)


try:
    import regex  # #

except ImportError:
    regex = None

import sre_constants
import sre_parse

categ_pattern = re.compile(r"\\p{[A-Za-z_]+}")


def get_regexp_width(expr):
    if regex:
        ##

        ##

        ##

        regexp_final = re.sub(categ_pattern, "A", expr)
    else:
        if re.search(categ_pattern, expr):
            raise ImportError(
                "`regex` module must be installed in order to use Unicode categories.",
                expr,
            )
        regexp_final = expr
    try:
        return [int(x) for x in sre_parse.parse(regexp_final).getwidth()]
    except sre_constants.error:
        if not regex:
            raise ValueError(expr)
        else:
            ##

            ##

            c = regex.compile(regexp_final)
            if c.match("") is None:
                ##

                return 1, int(sre_constants.MAXREPEAT)
            else:
                return 0, int(sre_constants.MAXREPEAT)


from collections import OrderedDict


class Meta:

    empty: bool
    line: int
    column: int
    start_pos: int
    end_line: int
    end_column: int
    end_pos: int
    orig_expansion: "List[TerminalDef]"
    match_tree: bool

    def __init__(self):
        self.empty = True


_Leaf_T = TypeVar("_Leaf_T")
Branch = Union[_Leaf_T, "Tree[_Leaf_T]"]


class Tree(Generic[_Leaf_T]):
    # --

    data: str
    children: "List[Branch[_Leaf_T]]"

    def __init__(
        self, data: str, children: "List[Branch[_Leaf_T]]", meta: Optional[Meta] = None
    ) -> None:
        self.data = data
        self.children = children
        self._meta = meta

    @property
    def meta(self) -> Meta:
        if self._meta is None:
            self._meta = Meta()
        return self._meta

    def __repr__(self):
        return "Tree(%r, %r)" % (self.data, self.children)

    def _pretty_label(self):
        return self.data

    def _pretty(self, level, indent_str):
        if len(self.children) == 1 and not isinstance(self.children[0], Tree):
            return [
                indent_str * level,
                self._pretty_label(),
                "\t",
                "%s" % (self.children[0],),
                "\n",
            ]

        l = [indent_str * level, self._pretty_label(), "\n"]
        for n in self.children:
            if isinstance(n, Tree):
                l += n._pretty(level + 1, indent_str)
            else:
                l += [indent_str * (level + 1), "%s" % (n,), "\n"]

        return l

    def pretty(self, indent_str: str = "  ") -> str:
        # --
        return "".join(self._pretty(0, indent_str))

    def __rich__(self, parent: "rich.tree.Tree" = None) -> "rich.tree.Tree":
        # --
        return self._rich(parent)

    def _rich(self, parent):
        if parent:
            tree = parent.add(f"[bold]{self.data}[/bold]")
        else:
            import rich.tree

            tree = rich.tree.Tree(self.data)

        for c in self.children:
            if isinstance(c, Tree):
                c._rich(tree)
            else:
                tree.add(f"[green]{c}[/green]")

        return tree

    def __eq__(self, other):
        try:
            return self.data == other.data and self.children == other.children
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self) -> int:
        return hash((self.data, tuple(self.children)))

    def iter_subtrees(self) -> "Iterator[Tree[_Leaf_T]]":
        # --
        queue = [self]
        subtrees = OrderedDict()
        for subtree in queue:
            subtrees[id(subtree)] = subtree
            ##

            queue += [
                c
                for c in reversed(subtree.children)  ##
                if isinstance(c, Tree) and id(c) not in subtrees
            ]

        del queue
        return reversed(list(subtrees.values()))

    def find_pred(
        self, pred: "Callable[[Tree[_Leaf_T]], bool]"
    ) -> "Iterator[Tree[_Leaf_T]]":
        # --
        return filter(pred, self.iter_subtrees())

    def find_data(self, data: str) -> "Iterator[Tree[_Leaf_T]]":
        # --
        return self.find_pred(lambda t: t.data == data)


from functools import update_wrapper, wraps
from inspect import getmembers, getmro

_Return_T = TypeVar("_Return_T")
_Return_V = TypeVar("_Return_V")
_Leaf_T = TypeVar("_Leaf_T")
_Leaf_U = TypeVar("_Leaf_U")
_R = TypeVar("_R")
_FUNC = Callable[..., _Return_T]
_DECORATED = Union[_FUNC, type]


class _DiscardType:
    # --

    def __repr__(self):
        return "lark.visitors.Discard"


Discard = _DiscardType()

##


class _Decoratable:
    # --

    @classmethod
    def _apply_v_args(cls, visit_wrapper):
        mro = getmro(cls)
        assert mro[0] is cls
        libmembers = {name for _cls in mro[1:] for name, _ in getmembers(_cls)}
        for name, value in getmembers(cls):

            ##

            if name.startswith("_") or (
                name in libmembers and name not in cls.__dict__
            ):
                continue
            if not callable(value):
                continue

            ##

            if isinstance(cls.__dict__[name], _VArgsWrapper):
                continue

            setattr(cls, name, _VArgsWrapper(cls.__dict__[name], visit_wrapper))
        return cls

    def __class_getitem__(cls, _):
        return cls


class Transformer(_Decoratable, ABC, Generic[_Leaf_T, _Return_T]):
    # --
    __visit_tokens__ = True  ##

    def __init__(self, visit_tokens: bool = True) -> None:
        self.__visit_tokens__ = visit_tokens

    def _call_userfunc(self, tree, new_children=None):
        ##

        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            try:
                wrapper = getattr(f, "visit_wrapper", None)
                if wrapper is not None:
                    return f.visit_wrapper(f, tree.data, children, tree.meta)
                else:
                    return f(children)
            except GrammarError:
                raise
            except Exception as e:
                raise VisitError(tree.data, tree, e)

    def _call_userfunc_token(self, token):
        try:
            f = getattr(self, token.type)
        except AttributeError:
            return self.__default_token__(token)
        else:
            try:
                return f(token)
            except GrammarError:
                raise
            except Exception as e:
                raise VisitError(token.type, token, e)

    def _transform_children(self, children):
        for c in children:
            if isinstance(c, Tree):
                res = self._transform_tree(c)
            elif self.__visit_tokens__ and isinstance(c, Token):
                res = self._call_userfunc_token(c)
            else:
                res = c

            if res is not Discard:
                yield res

    def _transform_tree(self, tree):
        children = list(self._transform_children(tree.children))
        return self._call_userfunc(tree, children)

    def transform(self, tree: Tree[_Leaf_T]) -> _Return_T:
        # --
        return self._transform_tree(tree)

    def __mul__(
        self: "Transformer[_Leaf_T, Tree[_Leaf_U]]",
        other: "Union[Transformer[_Leaf_U, _Return_V], TransformerChain[_Leaf_U, _Return_V,]]",
    ) -> "TransformerChain[_Leaf_T, _Return_V]":
        # --
        return TransformerChain(self, other)

    def __default__(self, data, children, meta):
        # --
        return Tree(data, children, meta)

    def __default_token__(self, token):
        # --
        return token


def merge_transformers(base_transformer=None, **transformers_to_merge):
    # --
    if base_transformer is None:
        base_transformer = Transformer()
    for prefix, transformer in transformers_to_merge.items():
        for method_name in dir(transformer):
            method = getattr(transformer, method_name)
            if not callable(method):
                continue
            if method_name.startswith("_") or method_name == "transform":
                continue
            prefixed_method = prefix + "__" + method_name
            if hasattr(base_transformer, prefixed_method):
                raise AttributeError(
                    "Cannot merge: method '%s' appears more than once" % prefixed_method
                )

            setattr(base_transformer, prefixed_method, method)

    return base_transformer


class InlineTransformer(Transformer):  ##
    def _call_userfunc(self, tree, new_children=None):
        ##

        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            return f(*children)


class TransformerChain(Generic[_Leaf_T, _Return_T]):

    transformers: "Tuple[Union[Transformer, TransformerChain], ...]"

    def __init__(self, *transformers: "Union[Transformer, TransformerChain]") -> None:
        self.transformers = transformers

    def transform(self, tree: Tree[_Leaf_T]) -> _Return_T:
        for t in self.transformers:
            tree = t.transform(tree)
        return cast(_Return_T, tree)

    def __mul__(
        self: "TransformerChain[_Leaf_T, Tree[_Leaf_U]]",
        other: "Union[Transformer[_Leaf_U, _Return_V], TransformerChain[_Leaf_U, _Return_V]]",
    ) -> "TransformerChain[_Leaf_T, _Return_V]":
        return TransformerChain(*self.transformers + (other,))


class Transformer_InPlace(Transformer):
    # --
    def _transform_tree(self, tree):  ##

        return self._call_userfunc(tree)

    def transform(self, tree: Tree[_Leaf_T]) -> _Return_T:
        for subtree in tree.iter_subtrees():
            subtree.children = list(self._transform_children(subtree.children))

        return self._transform_tree(tree)


class Transformer_NonRecursive(Transformer):
    # --

    def transform(self, tree: Tree[_Leaf_T]) -> _Return_T:
        ##

        rev_postfix = []
        q: List[Branch[_Leaf_T]] = [tree]
        while q:
            t = q.pop()
            rev_postfix.append(t)
            if isinstance(t, Tree):
                q += t.children

        ##

        stack: List = []
        for x in reversed(rev_postfix):
            if isinstance(x, Tree):
                size = len(x.children)
                if size:
                    args = stack[-size:]
                    del stack[-size:]
                else:
                    args = []

                res = self._call_userfunc(x, args)
                if res is not Discard:
                    stack.append(res)

            elif self.__visit_tokens__ and isinstance(x, Token):
                res = self._call_userfunc_token(x)
                if res is not Discard:
                    stack.append(res)
            else:
                stack.append(x)

        (result,) = stack  ##

        ##

        ##

        ##

        return cast(_Return_T, result)


class Transformer_InPlaceRecursive(Transformer):
    # --
    def _transform_tree(self, tree):
        tree.children = list(self._transform_children(tree.children))
        return self._call_userfunc(tree)


##


class VisitorBase:
    def _call_userfunc(self, tree):
        return getattr(self, tree.data, self.__default__)(tree)

    def __default__(self, tree):
        # --
        return tree

    def __class_getitem__(cls, _):
        return cls


class Visitor(VisitorBase, ABC, Generic[_Leaf_T]):
    # --

    def visit(self, tree: Tree[_Leaf_T]) -> Tree[_Leaf_T]:
        # --
        for subtree in tree.iter_subtrees():
            self._call_userfunc(subtree)
        return tree

    def visit_topdown(self, tree: Tree[_Leaf_T]) -> Tree[_Leaf_T]:
        # --
        for subtree in tree.iter_subtrees_topdown():
            self._call_userfunc(subtree)
        return tree


class Visitor_Recursive(VisitorBase, Generic[_Leaf_T]):
    # --

    def visit(self, tree: Tree[_Leaf_T]) -> Tree[_Leaf_T]:
        # --
        for child in tree.children:
            if isinstance(child, Tree):
                self.visit(child)

        self._call_userfunc(tree)
        return tree

    def visit_topdown(self, tree: Tree[_Leaf_T]) -> Tree[_Leaf_T]:
        # --
        self._call_userfunc(tree)

        for child in tree.children:
            if isinstance(child, Tree):
                self.visit_topdown(child)

        return tree


class Interpreter(_Decoratable, ABC, Generic[_Leaf_T, _Return_T]):
    # --

    def visit(self, tree: Tree[_Leaf_T]) -> _Return_T:
        ##

        ##

        ##

        return self._visit_tree(tree)

    def _visit_tree(self, tree: Tree[_Leaf_T]):
        f = getattr(self, tree.data)
        wrapper = getattr(f, "visit_wrapper", None)
        if wrapper is not None:
            return f.visit_wrapper(f, tree.data, tree.children, tree.meta)
        else:
            return f(tree)

    def visit_children(self, tree: Tree[_Leaf_T]) -> List:
        return [
            self._visit_tree(child) if isinstance(child, Tree) else child
            for child in tree.children
        ]

    def __getattr__(self, name):
        return self.__default__

    def __default__(self, tree):
        return self.visit_children(tree)


_InterMethod = Callable[[Type[Interpreter], _Return_T], _R]


def visit_children_decor(func: _InterMethod) -> _InterMethod:
    # --
    @wraps(func)
    def inner(cls, tree):
        values = cls.visit_children(tree)
        return func(cls, values)

    return inner


##


def _apply_v_args(obj, visit_wrapper):
    try:
        _apply = obj._apply_v_args
    except AttributeError:
        return _VArgsWrapper(obj, visit_wrapper)
    else:
        return _apply(visit_wrapper)


class _VArgsWrapper:
    # --
    base_func: Callable

    def __init__(
        self, func: Callable, visit_wrapper: Callable[[Callable, str, list, Any], Any]
    ):
        if isinstance(func, _VArgsWrapper):
            func = func.base_func
        ##

        self.base_func = func  ##

        self.visit_wrapper = visit_wrapper
        update_wrapper(self, func)

    def __call__(self, *args, **kwargs):
        return self.base_func(*args, **kwargs)

    def __get__(self, instance, owner=None):
        try:
            g = self.base_func.__get__
        except AttributeError:
            return self
        else:
            return _VArgsWrapper(g(instance, owner), self.visit_wrapper)

    def __set_name__(self, owner, name):
        try:
            f = self.base_func.__set_name__
        except AttributeError:
            return
        else:
            f(owner, name)


def _vargs_inline(f, _data, children, _meta):
    return f(*children)


def _vargs_meta_inline(f, _data, children, meta):
    return f(meta, *children)


def _vargs_meta(f, _data, children, meta):
    return f(meta, children)


def _vargs_tree(f, data, children, meta):
    return f(Tree(data, children, meta))


def v_args(
    inline: bool = False,
    meta: bool = False,
    tree: bool = False,
    wrapper: Optional[Callable] = None,
) -> Callable[[_DECORATED], _DECORATED]:
    # --
    if tree and (meta or inline):
        raise ValueError(
            "Visitor functions cannot combine 'tree' with 'meta' or 'inline'."
        )

    func = None
    if meta:
        if inline:
            func = _vargs_meta_inline
        else:
            func = _vargs_meta
    elif inline:
        func = _vargs_inline
    elif tree:
        func = _vargs_tree

    if wrapper is not None:
        if func is not None:
            raise ValueError(
                "Cannot use 'wrapper' along with 'tree', 'meta' or 'inline'."
            )
        func = wrapper

    def _visitor_args_dec(obj):
        return _apply_v_args(obj, func)

    return _visitor_args_dec


TOKEN_DEFAULT_PRIORITY = 0


class Symbol(Serialize):
    __slots__ = ("name",)

    name: str
    is_term: ClassVar[bool] = NotImplemented

    def __init__(self, name: str) -> None:
        self.name = name

    def __eq__(self, other):
        assert isinstance(other, Symbol), other
        return self.is_term == other.is_term and self.name == other.name

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.name)

    fullrepr = property(__repr__)

    def renamed(self, f):
        return type(self)(f(self.name))


class Terminal(Symbol):
    __serialize_fields__ = "name", "filter_out"

    is_term: ClassVar[bool] = True

    def __init__(self, name, filter_out=False):
        self.name = name
        self.filter_out = filter_out

    @property
    def fullrepr(self):
        return "%s(%r, %r)" % (type(self).__name__, self.name, self.filter_out)

    def renamed(self, f):
        return type(self)(f(self.name), self.filter_out)


class NonTerminal(Symbol):
    __serialize_fields__ = ("name",)

    is_term: ClassVar[bool] = False


class RuleOptions(Serialize):
    __serialize_fields__ = (
        "keep_all_tokens",
        "expand1",
        "priority",
        "template_source",
        "empty_indices",
    )

    keep_all_tokens: bool
    expand1: bool
    priority: Optional[int]
    template_source: Optional[str]
    empty_indices: Tuple[bool, ...]

    def __init__(
        self,
        keep_all_tokens: bool = False,
        expand1: bool = False,
        priority: Optional[int] = None,
        template_source: Optional[str] = None,
        empty_indices: Tuple[bool, ...] = (),
    ) -> None:
        self.keep_all_tokens = keep_all_tokens
        self.expand1 = expand1
        self.priority = priority
        self.template_source = template_source
        self.empty_indices = empty_indices

    def __repr__(self):
        return "RuleOptions(%r, %r, %r, %r)" % (
            self.keep_all_tokens,
            self.expand1,
            self.priority,
            self.template_source,
        )


class Rule(Serialize):
    # --
    __slots__ = ("origin", "expansion", "alias", "options", "order", "_hash")

    __serialize_fields__ = "origin", "expansion", "order", "alias", "options"
    __serialize_namespace__ = Terminal, NonTerminal, RuleOptions

    def __init__(self, origin, expansion, order=0, alias=None, options=None):
        self.origin = origin
        self.expansion = expansion
        self.alias = alias
        self.order = order
        self.options = options or RuleOptions()
        self._hash = hash((self.origin, tuple(self.expansion)))

    def _deserialize(self):
        self._hash = hash((self.origin, tuple(self.expansion)))

    def __str__(self):
        return "<%s : %s>" % (
            self.origin.name,
            " ".join(x.name for x in self.expansion),
        )

    def __repr__(self):
        return "Rule(%r, %r, %r, %r)" % (
            self.origin,
            self.expansion,
            self.alias,
            self.options,
        )

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if not isinstance(other, Rule):
            return False
        return self.origin == other.origin and self.expansion == other.expansion


from copy import copy


class Pattern(Serialize, ABC):

    value: str
    flags: Collection[str]
    raw: Optional[str]
    type: ClassVar[str]

    def __init__(
        self, value: str, flags: Collection[str] = (), raw: Optional[str] = None
    ) -> None:
        self.value = value
        self.flags = frozenset(flags)
        self.raw = raw

    def __repr__(self):
        return repr(self.to_regexp())

    ##

    def __hash__(self):
        return hash((type(self), self.value, self.flags))

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self.value == other.value
            and self.flags == other.flags
        )

    @abstractmethod
    def to_regexp(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def min_width(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def max_width(self) -> int:
        raise NotImplementedError()

    def _get_flags(self, value):
        for f in self.flags:
            value = "(?%s:%s)" % (f, value)
        return value


class PatternStr(Pattern):
    __serialize_fields__ = "value", "flags"

    type: ClassVar[str] = "str"

    def to_regexp(self) -> str:
        return self._get_flags(re.escape(self.value))

    @property
    def min_width(self) -> int:
        return len(self.value)

    @property
    def max_width(self) -> int:
        return len(self.value)


class PatternRE(Pattern):
    __serialize_fields__ = "value", "flags", "_width"

    type: ClassVar[str] = "re"

    def to_regexp(self) -> str:
        return self._get_flags(self.value)

    _width = None

    def _get_width(self):
        if self._width is None:
            self._width = get_regexp_width(self.to_regexp())
        return self._width

    @property
    def min_width(self) -> int:
        return self._get_width()[0]

    @property
    def max_width(self) -> int:
        return self._get_width()[1]


class TerminalDef(Serialize):
    __serialize_fields__ = "name", "pattern", "priority"
    __serialize_namespace__ = PatternStr, PatternRE

    name: str
    pattern: Pattern
    priority: int

    def __init__(
        self, name: str, pattern: Pattern, priority: int = TOKEN_DEFAULT_PRIORITY
    ) -> None:
        assert isinstance(pattern, Pattern), pattern
        self.name = name
        self.pattern = pattern
        self.priority = priority

    def __repr__(self):
        return "%s(%r, %r)" % (type(self).__name__, self.name, self.pattern)

    def user_repr(self) -> str:
        if self.name.startswith("__"):  ##

            return self.pattern.raw or self.name
        else:
            return self.name


_T = TypeVar("_T", bound="Token")


class Token(str):
    # --
    __slots__ = (
        "type",
        "start_pos",
        "value",
        "line",
        "column",
        "end_line",
        "end_column",
        "end_pos",
    )

    type: str
    start_pos: int
    value: Any
    line: int
    column: int
    end_line: int
    end_column: int
    end_pos: int

    def __new__(
        cls,
        type_,
        value,
        start_pos=None,
        line=None,
        column=None,
        end_line=None,
        end_column=None,
        end_pos=None,
    ):
        inst = super(Token, cls).__new__(cls, value)
        inst.type = type_
        inst.start_pos = start_pos
        inst.value = value
        inst.line = line
        inst.column = column
        inst.end_line = end_line
        inst.end_column = end_column
        inst.end_pos = end_pos
        return inst

    def update(
        self, type_: Optional[str] = None, value: Optional[Any] = None
    ) -> "Token":
        return Token.new_borrow_pos(
            type_ if type_ is not None else self.type,
            value if value is not None else self.value,
            self,
        )

    @classmethod
    def new_borrow_pos(cls: Type[_T], type_: str, value: Any, borrow_t: "Token") -> _T:
        return cls(
            type_,
            value,
            borrow_t.start_pos,
            borrow_t.line,
            borrow_t.column,
            borrow_t.end_line,
            borrow_t.end_column,
            borrow_t.end_pos,
        )

    def __reduce__(self):
        return (
            self.__class__,
            (self.type, self.value, self.start_pos, self.line, self.column),
        )

    def __repr__(self):
        return "Token(%r, %r)" % (self.type, self.value)

    def __deepcopy__(self, memo):
        return Token(self.type, self.value, self.start_pos, self.line, self.column)

    def __eq__(self, other):
        if isinstance(other, Token) and self.type != other.type:
            return False

        return str.__eq__(self, other)

    __hash__ = str.__hash__


class LineCounter:
    __slots__ = "char_pos", "line", "column", "line_start_pos", "newline_char"

    def __init__(self, newline_char):
        self.newline_char = newline_char
        self.char_pos = 0
        self.line = 1
        self.column = 1
        self.line_start_pos = 0

    def __eq__(self, other):
        if not isinstance(other, LineCounter):
            return NotImplemented

        return (
            self.char_pos == other.char_pos and self.newline_char == other.newline_char
        )

    def feed(self, token: Token, test_newline=True):
        # --
        if test_newline:
            newlines = token.count(self.newline_char)
            if newlines:
                self.line += newlines
                self.line_start_pos = (
                    self.char_pos + token.rindex(self.newline_char) + 1
                )

        self.char_pos += len(token)
        self.column = self.char_pos - self.line_start_pos + 1


class UnlessCallback:
    def __init__(self, scanner):
        self.scanner = scanner

    def __call__(self, t):
        res = self.scanner.match(t.value, 0)
        if res:
            _value, t.type = res
        return t


class CallChain:
    def __init__(self, callback1, callback2, cond):
        self.callback1 = callback1
        self.callback2 = callback2
        self.cond = cond

    def __call__(self, t):
        t2 = self.callback1(t)
        return self.callback2(t) if self.cond(t2) else t2


def _get_match(re_, regexp, s, flags):
    m = re_.match(regexp, s, flags)
    if m:
        return m.group(0)


def _create_unless(terminals, g_regex_flags, re_, use_bytes):
    tokens_by_type = classify(terminals, lambda t: type(t.pattern))
    assert len(tokens_by_type) <= 2, tokens_by_type.keys()
    embedded_strs = set()
    callback = {}
    for retok in tokens_by_type.get(PatternRE, []):
        unless = []
        for strtok in tokens_by_type.get(PatternStr, []):
            if strtok.priority != retok.priority:
                continue
            s = strtok.pattern.value
            if s == _get_match(re_, retok.pattern.to_regexp(), s, g_regex_flags):
                unless.append(strtok)
                if strtok.pattern.flags <= retok.pattern.flags:
                    embedded_strs.add(strtok)
        if unless:
            callback[retok.name] = UnlessCallback(
                Scanner(
                    unless, g_regex_flags, re_, match_whole=True, use_bytes=use_bytes
                )
            )

    new_terminals = [t for t in terminals if t not in embedded_strs]
    return new_terminals, callback


class Scanner:
    def __init__(self, terminals, g_regex_flags, re_, use_bytes, match_whole=False):
        self.terminals = terminals
        self.g_regex_flags = g_regex_flags
        self.re_ = re_
        self.use_bytes = use_bytes
        self.match_whole = match_whole

        self.allowed_types = {t.name for t in self.terminals}

        self._mres = self._build_mres(terminals, len(terminals))

    def _build_mres(self, terminals, max_size):
        ##

        ##

        ##

        postfix = "$" if self.match_whole else ""
        mres = []
        while terminals:
            pattern = "|".join(
                "(?P<%s>%s)" % (t.name, t.pattern.to_regexp() + postfix)
                for t in terminals[:max_size]
            )
            if self.use_bytes:
                pattern = pattern.encode("latin-1")
            try:
                mre = self.re_.compile(pattern, self.g_regex_flags)
            except AssertionError:  ##

                return self._build_mres(terminals, max_size // 2)

            mres.append((mre, {i: n for n, i in mre.groupindex.items()}))
            terminals = terminals[max_size:]
        return mres

    def match(self, text, pos):
        for mre, type_from_index in self._mres:
            m = mre.match(text, pos)
            if m:
                return m.group(0), type_from_index[m.lastindex]


def _regexp_has_newline(r: str):
    # --
    return (
        "\n" in r or "\\n" in r or "\\s" in r or "[^" in r or ("(?s" in r and "." in r)
    )


class LexerState:
    # --

    __slots__ = "text", "line_ctr", "last_token"

    def __init__(self, text, line_ctr=None, last_token=None):
        self.text = text
        self.line_ctr = line_ctr or LineCounter(
            b"\n" if isinstance(text, bytes) else "\n"
        )
        self.last_token = last_token

    def __eq__(self, other):
        if not isinstance(other, LexerState):
            return NotImplemented

        return (
            self.text is other.text
            and self.line_ctr == other.line_ctr
            and self.last_token == other.last_token
        )

    def __copy__(self):
        return type(self)(self.text, copy(self.line_ctr), self.last_token)


class LexerThread:
    # --

    def __init__(self, lexer: "Lexer", lexer_state: LexerState):
        self.lexer = lexer
        self.state = lexer_state

    @classmethod
    def from_text(cls, lexer: "Lexer", text: str):
        return cls(lexer, LexerState(text))

    def lex(self, parser_state):
        return self.lexer.lex(self.state, parser_state)

    def __copy__(self):
        return type(self)(self.lexer, copy(self.state))

    _Token = Token


_Callback = Callable[[Token], Token]


class Lexer(ABC):
    # --
    @abstractmethod
    def lex(self, lexer_state: LexerState, parser_state: Any) -> Iterator[Token]:
        return NotImplemented

    def make_lexer_state(self, text):
        # --
        return LexerState(text)


class BasicLexer(Lexer):

    terminals: Collection[TerminalDef]
    ignore_types: FrozenSet[str]
    newline_types: FrozenSet[str]
    user_callbacks: Dict[str, _Callback]
    callback: Dict[str, _Callback]
    re: ModuleType

    def __init__(self, conf: "LexerConf") -> None:
        terminals = list(conf.terminals)
        assert all(isinstance(t, TerminalDef) for t in terminals), terminals

        self.re = conf.re_module

        if not conf.skip_validation:
            ##

            for t in terminals:
                try:
                    self.re.compile(t.pattern.to_regexp(), conf.g_regex_flags)
                except self.re.error:
                    raise LexError("Cannot compile token %s: %s" % (t.name, t.pattern))

                if t.pattern.min_width == 0:
                    raise LexError(
                        "Lexer does not allow zero-width terminals. (%s: %s)"
                        % (t.name, t.pattern)
                    )

            if not (set(conf.ignore) <= {t.name for t in terminals}):
                raise LexError(
                    "Ignore terminals are not defined: %s"
                    % (set(conf.ignore) - {t.name for t in terminals})
                )

        ##

        self.newline_types = frozenset(
            t.name for t in terminals if _regexp_has_newline(t.pattern.to_regexp())
        )
        self.ignore_types = frozenset(conf.ignore)

        terminals.sort(
            key=lambda x: (
                -x.priority,
                -x.pattern.max_width,
                -len(x.pattern.value),
                x.name,
            )
        )
        self.terminals = terminals
        self.user_callbacks = conf.callbacks
        self.g_regex_flags = conf.g_regex_flags
        self.use_bytes = conf.use_bytes
        self.terminals_by_name = conf.terminals_by_name

        self._scanner = None

    def _build_scanner(self):
        terminals, self.callback = _create_unless(
            self.terminals, self.g_regex_flags, self.re, self.use_bytes
        )
        assert all(self.callback.values())

        for type_, f in self.user_callbacks.items():
            if type_ in self.callback:
                ##

                self.callback[type_] = CallChain(
                    self.callback[type_], f, lambda t: t.type == type_
                )
            else:
                self.callback[type_] = f

        self._scanner = Scanner(terminals, self.g_regex_flags, self.re, self.use_bytes)

    @property
    def scanner(self):
        if self._scanner is None:
            self._build_scanner()
        return self._scanner

    def match(self, text, pos):
        return self.scanner.match(text, pos)

    def lex(self, state: LexerState, parser_state: Any) -> Iterator[Token]:
        with suppress(EOFError):
            while True:
                yield self.next_token(state, parser_state)

    def next_token(self, lex_state: LexerState, parser_state: Any = None) -> Token:
        line_ctr = lex_state.line_ctr
        while line_ctr.char_pos < len(lex_state.text):
            res = self.match(lex_state.text, line_ctr.char_pos)
            if not res:
                allowed = self.scanner.allowed_types - self.ignore_types
                if not allowed:
                    allowed = {"<END-OF-FILE>"}
                raise UnexpectedCharacters(
                    lex_state.text,
                    line_ctr.char_pos,
                    line_ctr.line,
                    line_ctr.column,
                    allowed=allowed,
                    token_history=lex_state.last_token and [lex_state.last_token],
                    state=parser_state,
                    terminals_by_name=self.terminals_by_name,
                )

            value, type_ = res

            if type_ not in self.ignore_types:
                t = Token(
                    type_, value, line_ctr.char_pos, line_ctr.line, line_ctr.column
                )
                line_ctr.feed(value, type_ in self.newline_types)
                t.end_line = line_ctr.line
                t.end_column = line_ctr.column
                t.end_pos = line_ctr.char_pos
                if t.type in self.callback:
                    t = self.callback[t.type](t)
                    if not isinstance(t, Token):
                        raise LexError(
                            "Callbacks must return a token (returned %r)" % t
                        )
                lex_state.last_token = t
                return t
            else:
                if type_ in self.callback:
                    t2 = Token(
                        type_, value, line_ctr.char_pos, line_ctr.line, line_ctr.column
                    )
                    self.callback[type_](t2)
                line_ctr.feed(value, type_ in self.newline_types)

        ##

        raise EOFError(self)


class ContextualLexer(Lexer):

    lexers: Dict[str, BasicLexer]
    root_lexer: BasicLexer

    def __init__(
        self,
        conf: "LexerConf",
        states: Dict[str, Collection[str]],
        always_accept: Collection[str] = (),
    ) -> None:
        terminals = list(conf.terminals)
        terminals_by_name = conf.terminals_by_name

        trad_conf = copy(conf)
        trad_conf.terminals = terminals

        lexer_by_tokens: Dict[FrozenSet[str], BasicLexer] = {}
        self.lexers = {}
        for state, accepts in states.items():
            key = frozenset(accepts)
            try:
                lexer = lexer_by_tokens[key]
            except KeyError:
                accepts = set(accepts) | set(conf.ignore) | set(always_accept)
                lexer_conf = copy(trad_conf)
                lexer_conf.terminals = [
                    terminals_by_name[n] for n in accepts if n in terminals_by_name
                ]
                lexer = BasicLexer(lexer_conf)
                lexer_by_tokens[key] = lexer

            self.lexers[state] = lexer

        assert trad_conf.terminals is terminals
        self.root_lexer = BasicLexer(trad_conf)

    def lex(self, lexer_state: LexerState, parser_state: Any) -> Iterator[Token]:
        try:
            while True:
                lexer = self.lexers[parser_state.position]
                yield lexer.next_token(lexer_state, parser_state)
        except EOFError:
            pass
        except UnexpectedCharacters as e:
            ##

            ##

            try:
                last_token = lexer_state.last_token  ##

                token = self.root_lexer.next_token(lexer_state, parser_state)
                raise UnexpectedToken(
                    token,
                    e.allowed,
                    state=parser_state,
                    token_history=[last_token],
                    terminals_by_name=self.root_lexer.terminals_by_name,
                )
            except UnexpectedCharacters:
                raise e  ##


_ParserArgType: "TypeAlias" = 'Literal["earley", "lalr", "cyk", "auto"]'
_LexerArgType: "TypeAlias" = 'Union[Literal["auto", "basic", "contextual", "dynamic", "dynamic_complete"], Type[Lexer]]'
_Callback = Callable[[Token], Token]


class LexerConf(Serialize):
    __serialize_fields__ = (
        "terminals",
        "ignore",
        "g_regex_flags",
        "use_bytes",
        "lexer_type",
    )
    __serialize_namespace__ = (TerminalDef,)

    terminals: Collection[TerminalDef]
    re_module: ModuleType
    ignore: Collection[str]
    postlex: "Optional[PostLex]"
    callbacks: Dict[str, _Callback]
    g_regex_flags: int
    skip_validation: bool
    use_bytes: bool
    lexer_type: Optional[_LexerArgType]

    def __init__(
        self,
        terminals: Collection[TerminalDef],
        re_module: ModuleType,
        ignore: Collection[str] = (),
        postlex: "Optional[PostLex]" = None,
        callbacks: Optional[Dict[str, _Callback]] = None,
        g_regex_flags: int = 0,
        skip_validation: bool = False,
        use_bytes: bool = False,
    ):
        self.terminals = terminals
        self.terminals_by_name = {t.name: t for t in self.terminals}
        assert len(self.terminals) == len(self.terminals_by_name)
        self.ignore = ignore
        self.postlex = postlex
        self.callbacks = callbacks or {}
        self.g_regex_flags = g_regex_flags
        self.re_module = re_module
        self.skip_validation = skip_validation
        self.use_bytes = use_bytes
        self.lexer_type = None

    def _deserialize(self):
        self.terminals_by_name = {t.name: t for t in self.terminals}

    def __deepcopy__(self, memo=None):
        return type(self)(
            deepcopy(self.terminals, memo),
            self.re_module,
            deepcopy(self.ignore, memo),
            deepcopy(self.postlex, memo),
            deepcopy(self.callbacks, memo),
            deepcopy(self.g_regex_flags, memo),
            deepcopy(self.skip_validation, memo),
            deepcopy(self.use_bytes, memo),
        )


class ParserConf(Serialize):
    __serialize_fields__ = "rules", "start", "parser_type"

    def __init__(self, rules, callbacks, start):
        assert isinstance(start, list)
        self.rules = rules
        self.callbacks = callbacks
        self.start = start

        self.parser_type = None


from functools import partial, wraps
from itertools import product, repeat


class ExpandSingleChild:
    def __init__(self, node_builder):
        self.node_builder = node_builder

    def __call__(self, children):
        if len(children) == 1:
            return children[0]
        else:
            return self.node_builder(children)


class PropagatePositions:
    def __init__(self, node_builder, node_filter=None):
        self.node_builder = node_builder
        self.node_filter = node_filter

    def __call__(self, children):
        res = self.node_builder(children)

        if isinstance(res, Tree):
            ##

            ##

            ##

            ##

            res_meta = res.meta

            first_meta = self._pp_get_meta(children)
            if first_meta is not None:
                if not hasattr(res_meta, "line"):
                    ##

                    res_meta.line = getattr(
                        first_meta, "container_line", first_meta.line
                    )
                    res_meta.column = getattr(
                        first_meta, "container_column", first_meta.column
                    )
                    res_meta.start_pos = getattr(
                        first_meta, "container_start_pos", first_meta.start_pos
                    )
                    res_meta.empty = False

                res_meta.container_line = getattr(
                    first_meta, "container_line", first_meta.line
                )
                res_meta.container_column = getattr(
                    first_meta, "container_column", first_meta.column
                )

            last_meta = self._pp_get_meta(reversed(children))
            if last_meta is not None:
                if not hasattr(res_meta, "end_line"):
                    res_meta.end_line = getattr(
                        last_meta, "container_end_line", last_meta.end_line
                    )
                    res_meta.end_column = getattr(
                        last_meta, "container_end_column", last_meta.end_column
                    )
                    res_meta.end_pos = getattr(
                        last_meta, "container_end_pos", last_meta.end_pos
                    )
                    res_meta.empty = False

                res_meta.container_end_line = getattr(
                    last_meta, "container_end_line", last_meta.end_line
                )
                res_meta.container_end_column = getattr(
                    last_meta, "container_end_column", last_meta.end_column
                )

        return res

    def _pp_get_meta(self, children):
        for c in children:
            if self.node_filter is not None and not self.node_filter(c):
                continue
            if isinstance(c, Tree):
                if not c.meta.empty:
                    return c.meta
            elif isinstance(c, Token):
                return c


def make_propagate_positions(option):
    if callable(option):
        return partial(PropagatePositions, node_filter=option)
    elif option is True:
        return PropagatePositions
    elif option is False:
        return None

    raise ConfigurationError("Invalid option for propagate_positions: %r" % option)


class ChildFilter:
    def __init__(self, to_include, append_none, node_builder):
        self.node_builder = node_builder
        self.to_include = to_include
        self.append_none = append_none

    def __call__(self, children):
        filtered = []

        for i, to_expand, add_none in self.to_include:
            if add_none:
                filtered += [None] * add_none
            if to_expand:
                filtered += children[i].children
            else:
                filtered.append(children[i])

        if self.append_none:
            filtered += [None] * self.append_none

        return self.node_builder(filtered)


class ChildFilterLALR(ChildFilter):
    # --

    def __call__(self, children):
        filtered = []
        for i, to_expand, add_none in self.to_include:
            if add_none:
                filtered += [None] * add_none
            if to_expand:
                if filtered:
                    filtered += children[i].children
                else:  ##

                    filtered = children[i].children
            else:
                filtered.append(children[i])

        if self.append_none:
            filtered += [None] * self.append_none

        return self.node_builder(filtered)


class ChildFilterLALR_NoPlaceholders(ChildFilter):
    # --
    def __init__(self, to_include, node_builder):
        self.node_builder = node_builder
        self.to_include = to_include

    def __call__(self, children):
        filtered = []
        for i, to_expand in self.to_include:
            if to_expand:
                if filtered:
                    filtered += children[i].children
                else:  ##

                    filtered = children[i].children
            else:
                filtered.append(children[i])
        return self.node_builder(filtered)


def _should_expand(sym):
    return not sym.is_term and sym.name.startswith("_")


def maybe_create_child_filter(
    expansion, keep_all_tokens, ambiguous, _empty_indices: List[bool]
):
    ##

    if _empty_indices:
        assert _empty_indices.count(False) == len(expansion)
        s = "".join(str(int(b)) for b in _empty_indices)
        empty_indices = [len(ones) for ones in s.split("0")]
        assert len(empty_indices) == len(expansion) + 1, (empty_indices, len(expansion))
    else:
        empty_indices = [0] * (len(expansion) + 1)

    to_include = []
    nones_to_add = 0
    for i, sym in enumerate(expansion):
        nones_to_add += empty_indices[i]
        if keep_all_tokens or not (sym.is_term and sym.filter_out):
            to_include.append((i, _should_expand(sym), nones_to_add))
            nones_to_add = 0

    nones_to_add += empty_indices[len(expansion)]

    if (
        _empty_indices
        or len(to_include) < len(expansion)
        or any(to_expand for i, to_expand, _ in to_include)
    ):
        if _empty_indices or ambiguous:
            return partial(
                ChildFilter if ambiguous else ChildFilterLALR, to_include, nones_to_add
            )
        else:
            ##

            return partial(
                ChildFilterLALR_NoPlaceholders, [(i, x) for i, x, _ in to_include]
            )


class AmbiguousExpander:
    # --
    def __init__(self, to_expand, tree_class, node_builder):
        self.node_builder = node_builder
        self.tree_class = tree_class
        self.to_expand = to_expand

    def __call__(self, children):
        def _is_ambig_tree(t):
            return hasattr(t, "data") and t.data == "_ambig"

        ##

        ##

        ##

        ##

        ambiguous = []
        for i, child in enumerate(children):
            if _is_ambig_tree(child):
                if i in self.to_expand:
                    ambiguous.append(i)

                child.expand_kids_by_data("_ambig")

        if not ambiguous:
            return self.node_builder(children)

        expand = [
            iter(child.children) if i in ambiguous else repeat(child)
            for i, child in enumerate(children)
        ]
        return self.tree_class(
            "_ambig", [self.node_builder(list(f[0])) for f in product(zip(*expand))]
        )


def maybe_create_ambiguous_expander(tree_class, expansion, keep_all_tokens):
    to_expand = [
        i
        for i, sym in enumerate(expansion)
        if keep_all_tokens
        or ((not (sym.is_term and sym.filter_out)) and _should_expand(sym))
    ]
    if to_expand:
        return partial(AmbiguousExpander, to_expand, tree_class)


class AmbiguousIntermediateExpander:
    # --

    def __init__(self, tree_class, node_builder):
        self.node_builder = node_builder
        self.tree_class = tree_class

    def __call__(self, children):
        def _is_iambig_tree(child):
            return hasattr(child, "data") and child.data == "_iambig"

        def _collapse_iambig(children):
            # --

            ##

            ##

            if children and _is_iambig_tree(children[0]):
                iambig_node = children[0]
                result = []
                for grandchild in iambig_node.children:
                    collapsed = _collapse_iambig(grandchild.children)
                    if collapsed:
                        for child in collapsed:
                            child.children += children[1:]
                        result += collapsed
                    else:
                        new_tree = self.tree_class(
                            "_inter", grandchild.children + children[1:]
                        )
                        result.append(new_tree)
                return result

        collapsed = _collapse_iambig(children)
        if collapsed:
            processed_nodes = [self.node_builder(c.children) for c in collapsed]
            return self.tree_class("_ambig", processed_nodes)

        return self.node_builder(children)


def inplace_transformer(func):
    @wraps(func)
    def f(children):
        ##

        tree = Tree(func.__name__, children)
        return func(tree)

    return f


def apply_visit_wrapper(func, name, wrapper):
    if wrapper is _vargs_meta or wrapper is _vargs_meta_inline:
        raise NotImplementedError("Meta args not supported for internal transformer")

    @wraps(func)
    def f(children):
        return wrapper(func, name, children, None)

    return f


class ParseTreeBuilder:
    def __init__(
        self,
        rules,
        tree_class,
        propagate_positions=False,
        ambiguous=False,
        maybe_placeholders=False,
    ):
        self.tree_class = tree_class
        self.propagate_positions = propagate_positions
        self.ambiguous = ambiguous
        self.maybe_placeholders = maybe_placeholders

        self.rule_builders = list(self._init_builders(rules))

    def _init_builders(self, rules):
        propagate_positions = make_propagate_positions(self.propagate_positions)

        for rule in rules:
            options = rule.options
            keep_all_tokens = options.keep_all_tokens
            expand_single_child = options.expand1

            wrapper_chain = list(
                filter(
                    None,
                    [
                        (expand_single_child and not rule.alias) and ExpandSingleChild,
                        maybe_create_child_filter(
                            rule.expansion,
                            keep_all_tokens,
                            self.ambiguous,
                            options.empty_indices if self.maybe_placeholders else None,
                        ),
                        propagate_positions,
                        self.ambiguous
                        and maybe_create_ambiguous_expander(
                            self.tree_class, rule.expansion, keep_all_tokens
                        ),
                        self.ambiguous
                        and partial(AmbiguousIntermediateExpander, self.tree_class),
                    ],
                )
            )

            yield rule, wrapper_chain

    def create_callback(self, transformer=None):
        callbacks = {}

        default_handler = getattr(transformer, "__default__", None)
        if default_handler:

            def default_callback(data, children):
                return default_handler(data, children, None)

        else:
            default_callback = self.tree_class

        for rule, wrapper_chain in self.rule_builders:

            user_callback_name = (
                rule.alias or rule.options.template_source or rule.origin.name
            )
            try:
                f = getattr(transformer, user_callback_name)
                wrapper = getattr(f, "visit_wrapper", None)
                if wrapper is not None:
                    f = apply_visit_wrapper(f, user_callback_name, wrapper)
                elif isinstance(transformer, Transformer_InPlace):
                    f = inplace_transformer(f)
            except AttributeError:
                f = partial(default_callback, user_callback_name)

            for w in wrapper_chain:
                f = w(f)

            if rule in callbacks:
                raise GrammarError("Rule '%s' already exists" % (rule,))

            callbacks[rule] = f

        return callbacks


class LALR_Parser(Serialize):
    def __init__(self, parser_conf, debug=False):
        analysis = LALR_Analyzer(parser_conf, debug=debug)
        analysis.compute_lalr()
        callbacks = parser_conf.callbacks

        self._parse_table = analysis.parse_table
        self.parser_conf = parser_conf
        self.parser = _Parser(analysis.parse_table, callbacks, debug)

    @classmethod
    def deserialize(cls, data, memo, callbacks, debug=False):
        inst = cls.__new__(cls)
        inst._parse_table = IntParseTable.deserialize(data, memo)
        inst.parser = _Parser(inst._parse_table, callbacks, debug)
        return inst

    def serialize(self, memo):
        return self._parse_table.serialize(memo)

    def parse_interactive(self, lexer, start):
        return self.parser.parse(lexer, start, start_interactive=True)

    def parse(self, lexer, start, on_error=None):
        try:
            return self.parser.parse(lexer, start)
        except UnexpectedInput as e:
            if on_error is None:
                raise

            while True:
                if isinstance(e, UnexpectedCharacters):
                    s = e.interactive_parser.lexer_thread.state
                    p = s.line_ctr.char_pos

                if not on_error(e):
                    raise e

                if isinstance(e, UnexpectedCharacters):
                    ##

                    if p == s.line_ctr.char_pos:
                        s.line_ctr.feed(s.text[p : p + 1])

                try:
                    return e.interactive_parser.resume_parse()
                except UnexpectedToken as e2:
                    if (
                        isinstance(e, UnexpectedToken)
                        and e.token.type == e2.token.type == "$END"
                        and e.interactive_parser == e2.interactive_parser
                    ):
                        ##

                        raise e2
                    e = e2
                except UnexpectedCharacters as e2:
                    e = e2


class ParseConf:
    __slots__ = (
        "parse_table",
        "callbacks",
        "start",
        "start_state",
        "end_state",
        "states",
    )

    def __init__(self, parse_table, callbacks, start):
        self.parse_table = parse_table

        self.start_state = self.parse_table.start_states[start]
        self.end_state = self.parse_table.end_states[start]
        self.states = self.parse_table.states

        self.callbacks = callbacks
        self.start = start


class ParserState:
    __slots__ = "parse_conf", "lexer", "state_stack", "value_stack"

    def __init__(self, parse_conf, lexer, state_stack=None, value_stack=None):
        self.parse_conf = parse_conf
        self.lexer = lexer
        self.state_stack = state_stack or [self.parse_conf.start_state]
        self.value_stack = value_stack or []

    @property
    def position(self):
        return self.state_stack[-1]

    ##

    def __eq__(self, other):
        if not isinstance(other, ParserState):
            return NotImplemented
        return (
            len(self.state_stack) == len(other.state_stack)
            and self.position == other.position
        )

    def __copy__(self):
        return type(self)(
            self.parse_conf,
            self.lexer,  ##
            copy(self.state_stack),
            deepcopy(self.value_stack),
        )

    def copy(self):
        return copy(self)

    def feed_token(self, token, is_end=False):
        state_stack = self.state_stack
        value_stack = self.value_stack
        states = self.parse_conf.states
        end_state = self.parse_conf.end_state
        callbacks = self.parse_conf.callbacks

        while True:
            state = state_stack[-1]
            try:
                action, arg = states[state][token.type]
            except KeyError:
                expected = {s for s in states[state].keys() if s.isupper()}
                raise UnexpectedToken(
                    token, expected, state=self, interactive_parser=None
                )

            assert arg != end_state

            if action is Shift:
                ##

                assert not is_end
                state_stack.append(arg)
                value_stack.append(
                    token
                    if token.type not in callbacks
                    else callbacks[token.type](token)
                )
                return
            else:
                ##

                rule = arg
                size = len(rule.expansion)
                if size:
                    s = value_stack[-size:]
                    del state_stack[-size:]
                    del value_stack[-size:]
                else:
                    s = []

                value = callbacks[rule](s)

                _action, new_state = states[state_stack[-1]][rule.origin.name]
                assert _action is Shift
                state_stack.append(new_state)
                value_stack.append(value)

                if is_end and state_stack[-1] == end_state:
                    return value_stack[-1]


class _Parser:
    def __init__(self, parse_table, callbacks, debug=False):
        self.parse_table = parse_table
        self.callbacks = callbacks
        self.debug = debug

    def parse(
        self, lexer, start, value_stack=None, state_stack=None, start_interactive=False
    ):
        parse_conf = ParseConf(self.parse_table, self.callbacks, start)
        parser_state = ParserState(parse_conf, lexer, state_stack, value_stack)
        if start_interactive:
            return InteractiveParser(self, parser_state, parser_state.lexer)
        return self.parse_from_state(parser_state)

    def parse_from_state(self, state):
        ##

        try:
            token = None
            for token in state.lexer.lex(state):
                state.feed_token(token)

            end_token = (
                Token.new_borrow_pos("$END", "", token)
                if token
                else Token("$END", "", 0, 1, 1)
            )
            return state.feed_token(end_token, True)
        except UnexpectedInput as e:
            try:
                e.interactive_parser = InteractiveParser(self, state, state.lexer)
            except NameError:
                pass
            raise e
        except Exception as e:
            if self.debug:
                print("")
                print("STATE STACK DUMP")
                print("----------------")
                for i, s in enumerate(state.state_stack):
                    print("%d)" % i, s)
                print("")

            raise


class Action:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)


Shift = Action("Shift")
Reduce = Action("Reduce")


class ParseTable:
    def __init__(self, states, start_states, end_states):
        self.states = states
        self.start_states = start_states
        self.end_states = end_states

    def serialize(self, memo):
        tokens = Enumerator()
        rules = Enumerator()

        states = {
            state: {
                tokens.get(token): (
                    (1, arg.serialize(memo)) if action is Reduce else (0, arg)
                )
                for token, (action, arg) in actions.items()
            }
            for state, actions in self.states.items()
        }

        return {
            "tokens": tokens.reversed(),
            "states": states,
            "start_states": self.start_states,
            "end_states": self.end_states,
        }

    @classmethod
    def deserialize(cls, data, memo):
        tokens = data["tokens"]
        states = {
            state: {
                tokens[token]: (
                    (Reduce, Rule.deserialize(arg, memo))
                    if action == 1
                    else (Shift, arg)
                )
                for token, (action, arg) in actions.items()
            }
            for state, actions in data["states"].items()
        }
        return cls(states, data["start_states"], data["end_states"])


class IntParseTable(ParseTable):
    @classmethod
    def from_ParseTable(cls, parse_table):
        enum = list(parse_table.states)
        state_to_idx = {s: i for i, s in enumerate(enum)}
        int_states = {}

        for s, la in parse_table.states.items():
            la = {
                k: (v[0], state_to_idx[v[1]]) if v[0] is Shift else v
                for k, v in la.items()
            }
            int_states[state_to_idx[s]] = la

        start_states = {
            start: state_to_idx[s] for start, s in parse_table.start_states.items()
        }
        end_states = {
            start: state_to_idx[s] for start, s in parse_table.end_states.items()
        }
        return cls(int_states, start_states, end_states)


def _wrap_lexer(lexer_class):
    future_interface = getattr(lexer_class, "__future_interface__", False)
    if future_interface:
        return lexer_class
    else:

        class CustomLexerWrapper(Lexer):
            def __init__(self, lexer_conf):
                self.lexer = lexer_class(lexer_conf)

            def lex(self, lexer_state, parser_state):
                return self.lexer.lex(lexer_state.text)

        return CustomLexerWrapper


def _deserialize_parsing_frontend(data, memo, lexer_conf, callbacks, options):
    parser_conf = ParserConf.deserialize(data["parser_conf"], memo)
    cls = (options and options._plugins.get("LALR_Parser")) or LALR_Parser
    parser = cls.deserialize(data["parser"], memo, callbacks, options.debug)
    parser_conf.callbacks = callbacks
    return ParsingFrontend(lexer_conf, parser_conf, options, parser=parser)


_parser_creators: "Dict[str, Callable[[LexerConf, Any, Any], Any]]" = {}


class ParsingFrontend(Serialize):
    __serialize_fields__ = "lexer_conf", "parser_conf", "parser"

    def __init__(self, lexer_conf, parser_conf, options, parser=None):
        self.parser_conf = parser_conf
        self.lexer_conf = lexer_conf
        self.options = options

        ##

        if parser:  ##

            self.parser = parser
        else:
            create_parser = _parser_creators.get(parser_conf.parser_type)
            assert (
                create_parser is not None
            ), "{} is not supported in standalone mode".format(parser_conf.parser_type)
            self.parser = create_parser(lexer_conf, parser_conf, options)

        ##

        lexer_type = lexer_conf.lexer_type
        self.skip_lexer = False
        if lexer_type in ("dynamic", "dynamic_complete"):
            assert lexer_conf.postlex is None
            self.skip_lexer = True
            return

        try:
            create_lexer = {
                "basic": create_basic_lexer,
                "contextual": create_contextual_lexer,
            }[lexer_type]
        except KeyError:
            assert issubclass(lexer_type, Lexer), lexer_type
            self.lexer = _wrap_lexer(lexer_type)(lexer_conf)
        else:
            self.lexer = create_lexer(
                lexer_conf, self.parser, lexer_conf.postlex, options
            )

        if lexer_conf.postlex:
            self.lexer = PostLexConnector(self.lexer, lexer_conf.postlex)

    def _verify_start(self, start=None):
        if start is None:
            start_decls = self.parser_conf.start
            if len(start_decls) > 1:
                raise ConfigurationError(
                    "Lark initialized with more than 1 possible start rule. Must specify which start rule to parse",
                    start_decls,
                )
            (start,) = start_decls
        elif start not in self.parser_conf.start:
            raise ConfigurationError(
                "Unknown start rule %s. Must be one of %r"
                % (start, self.parser_conf.start)
            )
        return start

    def _make_lexer_thread(self, text):
        cls = (self.options and self.options._plugins.get("LexerThread")) or LexerThread
        return text if self.skip_lexer else cls.from_text(self.lexer, text)

    def parse(self, text, start=None, on_error=None):
        chosen_start = self._verify_start(start)
        kw = {} if on_error is None else {"on_error": on_error}
        stream = self._make_lexer_thread(text)
        return self.parser.parse(stream, chosen_start, **kw)

    def parse_interactive(self, text=None, start=None):
        chosen_start = self._verify_start(start)
        if self.parser_conf.parser_type != "lalr":
            raise ConfigurationError(
                "parse_interactive() currently only works with parser='lalr' "
            )
        stream = self._make_lexer_thread(text)
        return self.parser.parse_interactive(stream, chosen_start)


def _validate_frontend_args(parser, lexer) -> None:
    assert_config(parser, ("lalr", "earley", "cyk"))
    if not isinstance(lexer, type):  ##

        expected = {
            "lalr": ("basic", "contextual"),
            "earley": ("basic", "dynamic", "dynamic_complete"),
            "cyk": ("basic",),
        }[parser]
        assert_config(
            lexer,
            expected,
            "Parser %r does not support lexer %%r, expected one of %%s" % parser,
        )


def _get_lexer_callbacks(transformer, terminals):
    result = {}
    for terminal in terminals:
        callback = getattr(transformer, terminal.name, None)
        if callback is not None:
            result[terminal.name] = callback
    return result


class PostLexConnector:
    def __init__(self, lexer, postlexer):
        self.lexer = lexer
        self.postlexer = postlexer

    def lex(self, lexer_state, parser_state):
        i = self.lexer.lex(lexer_state, parser_state)
        return self.postlexer.process(i)


def create_basic_lexer(lexer_conf, parser, postlex, options):
    cls = (options and options._plugins.get("BasicLexer")) or BasicLexer
    return cls(lexer_conf)


def create_contextual_lexer(lexer_conf, parser, postlex, options):
    cls = (options and options._plugins.get("ContextualLexer")) or ContextualLexer
    states = {idx: list(t.keys()) for idx, t in parser._parse_table.states.items()}
    always_accept = postlex.always_accept if postlex else ()
    return cls(lexer_conf, states, always_accept=always_accept)


def create_lalr_parser(lexer_conf, parser_conf, options=None):
    debug = options.debug if options else False
    cls = (options and options._plugins.get("LALR_Parser")) or LALR_Parser
    return cls(parser_conf, debug=debug)


_parser_creators["lalr"] = create_lalr_parser


class PostLex(ABC):
    @abstractmethod
    def process(self, stream: Iterator[Token]) -> Iterator[Token]:
        return stream

    always_accept: Iterable[str] = ()


class LarkOptions(Serialize):
    # --

    start: List[str]
    debug: bool
    transformer: "Optional[Transformer]"
    propagate_positions: Union[bool, str]
    maybe_placeholders: bool
    cache: Union[bool, str]
    regex: bool
    g_regex_flags: int
    keep_all_tokens: bool
    tree_class: Any
    parser: _ParserArgType
    lexer: _LexerArgType
    ambiguity: 'Literal["auto", "resolve", "explicit", "forest"]'
    postlex: Optional[PostLex]
    priority: 'Optional[Literal["auto", "normal", "invert"]]'
    lexer_callbacks: Dict[str, Callable[[Token], Token]]
    use_bytes: bool
    edit_terminals: Optional[Callable[[TerminalDef], TerminalDef]]
    import_paths: "List[Union[str, Callable[[Union[None, str, PackageResource], str], Tuple[str, str]]]]"
    source_path: Optional[str]

    OPTIONS_DOC = """
    **===  General Options  ===**

    start
            The start symbol. Either a string, or a list of strings for multiple possible starts (Default: "start")
    debug
            Display debug information and extra warnings. Use only when debugging (Default: ``False``)
            When used with Earley, it generates a forest graph as "sppf.png", if 'dot' is installed.
    transformer
            Applies the transformer to every parse tree (equivalent to applying it after the parse, but faster)
    propagate_positions
            Propagates (line, column, end_line, end_column) attributes into all tree branches.
            Accepts ``False``, ``True``, or a callable, which will filter which nodes to ignore when propagating.
    maybe_placeholders
            When ``True``, the ``[]`` operator returns ``None`` when not matched.
            When ``False``,  ``[]`` behaves like the ``?`` operator, and returns no value at all.
            (default= ``True``)
    cache
            Cache the results of the Lark grammar analysis, for x2 to x3 faster loading. LALR only for now.

            - When ``False``, does nothing (default)
            - When ``True``, caches to a temporary file in the local directory
            - When given a string, caches to the path pointed by the string
    regex
            When True, uses the ``regex`` module instead of the stdlib ``re``.
    g_regex_flags
            Flags that are applied to all terminals (both regex and strings)
    keep_all_tokens
            Prevent the tree builder from automagically removing "punctuation" tokens (Default: ``False``)
    tree_class
            Lark will produce trees comprised of instances of this class instead of the default ``lark.Tree``.

    **=== Algorithm Options ===**

    parser
            Decides which parser engine to use. Accepts "earley" or "lalr". (Default: "earley").
            (there is also a "cyk" option for legacy)
    lexer
            Decides whether or not to use a lexer stage

            - "auto" (default): Choose for me based on the parser
            - "basic": Use a basic lexer
            - "contextual": Stronger lexer (only works with parser="lalr")
            - "dynamic": Flexible and powerful (only with parser="earley")
            - "dynamic_complete": Same as dynamic, but tries *every* variation of tokenizing possible.
    ambiguity
            Decides how to handle ambiguity in the parse. Only relevant if parser="earley"

            - "resolve": The parser will automatically choose the simplest derivation
              (it chooses consistently: greedy for tokens, non-greedy for rules)
            - "explicit": The parser will return all derivations wrapped in "_ambig" tree nodes (i.e. a forest).
            - "forest": The parser will return the root of the shared packed parse forest.

    **=== Misc. / Domain Specific Options ===**

    postlex
            Lexer post-processing (Default: ``None``) Only works with the basic and contextual lexers.
    priority
            How priorities should be evaluated - "auto", ``None``, "normal", "invert" (Default: "auto")
    lexer_callbacks
            Dictionary of callbacks for the lexer. May alter tokens during lexing. Use with caution.
    use_bytes
            Accept an input of type ``bytes`` instead of ``str``.
    edit_terminals
            A callback for editing the terminals before parse.
    import_paths
            A List of either paths or loader functions to specify from where grammars are imported
    source_path
            Override the source of from where the grammar was loaded. Useful for relative imports and unconventional grammar loading
    **=== End of Options ===**
    """
    if __doc__:
        __doc__ += OPTIONS_DOC

    ##

    ##

    ##

    ##

    ##

    ##

    _defaults: Dict[str, Any] = {
        "debug": False,
        "keep_all_tokens": False,
        "tree_class": None,
        "cache": False,
        "postlex": None,
        "parser": "earley",
        "lexer": "auto",
        "transformer": None,
        "start": "start",
        "priority": "auto",
        "ambiguity": "auto",
        "regex": False,
        "propagate_positions": False,
        "lexer_callbacks": {},
        "maybe_placeholders": True,
        "edit_terminals": None,
        "g_regex_flags": 0,
        "use_bytes": False,
        "import_paths": [],
        "source_path": None,
        "_plugins": {},
    }

    def __init__(self, options_dict):
        o = dict(options_dict)

        options = {}
        for name, default in self._defaults.items():
            if name in o:
                value = o.pop(name)
                if isinstance(default, bool) and name not in (
                    "cache",
                    "use_bytes",
                    "propagate_positions",
                ):
                    value = bool(value)
            else:
                value = default

            options[name] = value

        if isinstance(options["start"], str):
            options["start"] = [options["start"]]

        self.__dict__["options"] = options

        assert_config(self.parser, ("earley", "lalr", "cyk", None))

        if self.parser == "earley" and self.transformer:
            raise ConfigurationError(
                "Cannot specify an embedded transformer when using the Earley algorithm. "
                "Please use your transformer on the resulting parse tree, or use a different algorithm (i.e. LALR)"
            )

        if o:
            raise ConfigurationError("Unknown options: %s" % o.keys())

    def __getattr__(self, name):
        try:
            return self.__dict__["options"][name]
        except KeyError as e:
            raise AttributeError(e)

    def __setattr__(self, name, value):
        assert_config(
            name, self.options.keys(), "%r isn't a valid option. Expected one of: %s"
        )
        self.options[name] = value

    def serialize(self, memo):
        return self.options

    @classmethod
    def deserialize(cls, data, memo):
        return cls(data)


##

##

_LOAD_ALLOWED_OPTIONS = {
    "postlex",
    "transformer",
    "lexer_callbacks",
    "use_bytes",
    "debug",
    "g_regex_flags",
    "regex",
    "propagate_positions",
    "tree_class",
    "_plugins",
}

_VALID_PRIORITY_OPTIONS = ("auto", "normal", "invert", None)
_VALID_AMBIGUITY_OPTIONS = ("auto", "resolve", "explicit", "forest")


_T = TypeVar("_T", bound="Lark")


class Lark(Serialize):
    # --

    source_path: str
    source_grammar: str
    grammar: "Grammar"
    options: LarkOptions
    lexer: Lexer
    terminals: List[TerminalDef]

    def __init__(self, grammar: "Union[Grammar, str, IO[str]]", **options) -> None:
        self.options = LarkOptions(options)

        ##

        use_regex = self.options.regex
        if use_regex:
            if regex:
                re_module = regex
            else:
                raise ImportError(
                    "`regex` module must be installed if calling `Lark(regex=True)`."
                )
        else:
            re_module = re

        ##

        if self.options.source_path is None:
            try:
                self.source_path = grammar.name
            except AttributeError:
                self.source_path = "<string>"
        else:
            self.source_path = self.options.source_path

        ##

        try:
            read = grammar.read
        except AttributeError:
            pass
        else:
            grammar = read()

        cache_fn = None
        cache_md5 = None
        if isinstance(grammar, str):
            self.source_grammar = grammar
            if self.options.use_bytes:
                if not isascii(grammar):
                    raise ConfigurationError(
                        "Grammar must be ascii only, when use_bytes=True"
                    )

            if self.options.cache:
                if self.options.parser != "lalr":
                    raise ConfigurationError(
                        "cache only works with parser='lalr' for now"
                    )

                unhashable = (
                    "transformer",
                    "postlex",
                    "lexer_callbacks",
                    "edit_terminals",
                    "_plugins",
                )
                options_str = "".join(
                    k + str(v) for k, v in options.items() if k not in unhashable
                )
                from . import __version__

                s = grammar + options_str + __version__ + str(sys.version_info[:2])
                cache_md5 = hashlib.md5(s.encode("utf8")).hexdigest()

                if isinstance(self.options.cache, str):
                    cache_fn = self.options.cache
                else:
                    if self.options.cache is not True:
                        raise ConfigurationError("cache argument must be bool or str")

                    cache_fn = tempfile.gettempdir() + "/.lark_cache_%s_%s_%s.tmp" % (
                        cache_md5,
                        *sys.version_info[:2],
                    )

                if FS.exists(cache_fn):
                    logger.debug("Loading grammar from cache: %s", cache_fn)
                    ##

                    for name in set(options) - _LOAD_ALLOWED_OPTIONS:
                        del options[name]
                    with FS.open(cache_fn, "rb") as f:
                        old_options = self.options
                        try:
                            file_md5 = f.readline().rstrip(b"\n")
                            cached_used_files = pickle.load(f)
                            if file_md5 == cache_md5.encode(
                                "utf8"
                            ) and verify_used_files(cached_used_files):
                                cached_parser_data = pickle.load(f)
                                self._load(cached_parser_data, **options)
                                return
                        except Exception:  ##

                            logger.exception(
                                "Failed to load Lark from cache: %r. We will try to carry on."
                                % cache_fn
                            )

                            ##

                            ##

                            self.options = old_options

            ##

            self.grammar, used_files = load_grammar(
                grammar,
                self.source_path,
                self.options.import_paths,
                self.options.keep_all_tokens,
            )
        else:
            assert isinstance(grammar, Grammar)
            self.grammar = grammar

        if self.options.lexer == "auto":
            if self.options.parser == "lalr":
                self.options.lexer = "contextual"
            elif self.options.parser == "earley":
                if self.options.postlex is not None:
                    logger.info(
                        "postlex can't be used with the dynamic lexer, so we use 'basic' instead. "
                        "Consider using lalr with contextual instead of earley"
                    )
                    self.options.lexer = "basic"
                else:
                    self.options.lexer = "dynamic"
            elif self.options.parser == "cyk":
                self.options.lexer = "basic"
            else:
                assert False, self.options.parser
        lexer = self.options.lexer
        if isinstance(lexer, type):
            assert issubclass(lexer, Lexer)  ##

        else:
            assert_config(lexer, ("basic", "contextual", "dynamic", "dynamic_complete"))
            if self.options.postlex is not None and "dynamic" in lexer:
                raise ConfigurationError(
                    "Can't use postlex with a dynamic lexer. Use basic or contextual instead"
                )

        if self.options.ambiguity == "auto":
            if self.options.parser == "earley":
                self.options.ambiguity = "resolve"
        else:
            assert_config(
                self.options.parser,
                ("earley", "cyk"),
                "%r doesn't support disambiguation. Use one of these parsers instead: %s",
            )

        if self.options.priority == "auto":
            self.options.priority = "normal"

        if self.options.priority not in _VALID_PRIORITY_OPTIONS:
            raise ConfigurationError(
                "invalid priority option: %r. Must be one of %r"
                % (self.options.priority, _VALID_PRIORITY_OPTIONS)
            )
        if self.options.ambiguity not in _VALID_AMBIGUITY_OPTIONS:
            raise ConfigurationError(
                "invalid ambiguity option: %r. Must be one of %r"
                % (self.options.ambiguity, _VALID_AMBIGUITY_OPTIONS)
            )

        if self.options.parser is None:
            terminals_to_keep = "*"
        elif self.options.postlex is not None:
            terminals_to_keep = set(self.options.postlex.always_accept)
        else:
            terminals_to_keep = set()

        ##

        self.terminals, self.rules, self.ignore_tokens = self.grammar.compile(
            self.options.start, terminals_to_keep
        )

        if self.options.edit_terminals:
            for t in self.terminals:
                self.options.edit_terminals(t)

        self._terminals_dict = {t.name: t for t in self.terminals}

        ##

        if self.options.priority == "invert":
            for rule in self.rules:
                if rule.options.priority is not None:
                    rule.options.priority = -rule.options.priority
            for term in self.terminals:
                term.priority = -term.priority
        ##

        ##

        ##

        elif self.options.priority is None:
            for rule in self.rules:
                if rule.options.priority is not None:
                    rule.options.priority = None
            for term in self.terminals:
                term.priority = 0

        ##

        self.lexer_conf = LexerConf(
            self.terminals,
            re_module,
            self.ignore_tokens,
            self.options.postlex,
            self.options.lexer_callbacks,
            self.options.g_regex_flags,
            use_bytes=self.options.use_bytes,
        )

        if self.options.parser:
            self.parser = self._build_parser()
        elif lexer:
            self.lexer = self._build_lexer()

        if cache_fn:
            logger.debug("Saving grammar to cache: %s", cache_fn)
            with FS.open(cache_fn, "wb") as f:
                assert cache_md5 is not None
                f.write(cache_md5.encode("utf8") + b"\n")
                pickle.dump(used_files, f)
                self.save(f, _LOAD_ALLOWED_OPTIONS)

    if __doc__:
        __doc__ += "\n\n" + LarkOptions.OPTIONS_DOC

    __serialize_fields__ = "parser", "rules", "options"

    def _build_lexer(self, dont_ignore=False):
        lexer_conf = self.lexer_conf
        if dont_ignore:
            from copy import copy

            lexer_conf = copy(lexer_conf)
            lexer_conf.ignore = ()
        return BasicLexer(lexer_conf)

    def _prepare_callbacks(self):
        self._callbacks = {}
        ##

        if self.options.ambiguity != "forest":
            self._parse_tree_builder = ParseTreeBuilder(
                self.rules,
                self.options.tree_class or Tree,
                self.options.propagate_positions,
                self.options.parser != "lalr" and self.options.ambiguity == "explicit",
                self.options.maybe_placeholders,
            )
            self._callbacks = self._parse_tree_builder.create_callback(
                self.options.transformer
            )
        self._callbacks.update(
            _get_lexer_callbacks(self.options.transformer, self.terminals)
        )

    def _build_parser(self):
        self._prepare_callbacks()
        _validate_frontend_args(self.options.parser, self.options.lexer)
        parser_conf = ParserConf(self.rules, self._callbacks, self.options.start)
        return _construct_parsing_frontend(
            self.options.parser,
            self.options.lexer,
            self.lexer_conf,
            parser_conf,
            options=self.options,
        )

    def save(self, f, exclude_options: Collection[str] = ()):
        # --
        data, m = self.memo_serialize([TerminalDef, Rule])
        if exclude_options:
            data["options"] = {
                n: v for n, v in data["options"].items() if n not in exclude_options
            }
        pickle.dump({"data": data, "memo": m}, f, protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load(cls, f):
        # --
        inst = cls.__new__(cls)
        return inst._load(f)

    def _deserialize_lexer_conf(self, data, memo, options):
        lexer_conf = LexerConf.deserialize(data["lexer_conf"], memo)
        lexer_conf.callbacks = options.lexer_callbacks or {}
        lexer_conf.re_module = regex if options.regex else re
        lexer_conf.use_bytes = options.use_bytes
        lexer_conf.g_regex_flags = options.g_regex_flags
        lexer_conf.skip_validation = True
        lexer_conf.postlex = options.postlex
        return lexer_conf

    def _load(self, f, **kwargs):
        if isinstance(f, dict):
            d = f
        else:
            d = pickle.load(f)
        memo_json = d["memo"]
        data = d["data"]

        assert memo_json
        memo = SerializeMemoizer.deserialize(
            memo_json, {"Rule": Rule, "TerminalDef": TerminalDef}, {}
        )
        options = dict(data["options"])
        if (set(kwargs) - _LOAD_ALLOWED_OPTIONS) & set(LarkOptions._defaults):
            raise ConfigurationError(
                "Some options are not allowed when loading a Parser: {}".format(
                    set(kwargs) - _LOAD_ALLOWED_OPTIONS
                )
            )
        options.update(kwargs)
        self.options = LarkOptions.deserialize(options, memo)
        self.rules = [Rule.deserialize(r, memo) for r in data["rules"]]
        self.source_path = "<deserialized>"
        _validate_frontend_args(self.options.parser, self.options.lexer)
        self.lexer_conf = self._deserialize_lexer_conf(
            data["parser"], memo, self.options
        )
        self.terminals = self.lexer_conf.terminals
        self._prepare_callbacks()
        self._terminals_dict = {t.name: t for t in self.terminals}
        self.parser = _deserialize_parsing_frontend(
            data["parser"],
            memo,
            self.lexer_conf,
            self._callbacks,
            self.options,  ##
        )
        return self

    @classmethod
    def _load_from_dict(cls, data, memo, **kwargs):
        inst = cls.__new__(cls)
        return inst._load({"data": data, "memo": memo}, **kwargs)

    @classmethod
    def open(
        cls: Type[_T], grammar_filename: str, rel_to: Optional[str] = None, **options
    ) -> _T:
        # --
        if rel_to:
            basepath = os.path.dirname(rel_to)
            grammar_filename = os.path.join(basepath, grammar_filename)
        with open(grammar_filename, encoding="utf8") as f:
            return cls(f, **options)

    @classmethod
    def open_from_package(
        cls: Type[_T],
        package: str,
        grammar_path: str,
        search_paths: "Sequence[str]" = [""],
        **options,
    ) -> _T:
        # --
        package_loader = FromPackageLoader(package, search_paths)
        full_path, text = package_loader(None, grammar_path)
        options.setdefault("source_path", full_path)
        options.setdefault("import_paths", [])
        options["import_paths"].append(package_loader)
        return cls(text, **options)

    def __repr__(self):
        return "Lark(open(%r), parser=%r, lexer=%r, ...)" % (
            self.source_path,
            self.options.parser,
            self.options.lexer,
        )

    def lex(self, text: str, dont_ignore: bool = False) -> Iterator[Token]:
        # --
        if not hasattr(self, "lexer") or dont_ignore:
            lexer = self._build_lexer(dont_ignore)
        else:
            lexer = self.lexer
        lexer_thread = LexerThread.from_text(lexer, text)
        stream = lexer_thread.lex(None)
        if self.options.postlex:
            return self.options.postlex.process(stream)
        return stream

    def get_terminal(self, name: str) -> TerminalDef:
        # --
        return self._terminals_dict[name]

    def parse_interactive(
        self, text: Optional[str] = None, start: Optional[str] = None
    ) -> "InteractiveParser":
        # --
        return self.parser.parse_interactive(text, start=start)

    def parse(
        self,
        text: str,
        start: Optional[str] = None,
        on_error: "Optional[Callable[[UnexpectedInput], bool]]" = None,
    ) -> "ParseTree":
        # --
        return self.parser.parse(text, start=start, on_error=on_error)


class DedentError(LarkError):
    pass


class Indenter(PostLex, ABC):
    paren_level: int
    indent_level: List[int]

    def __init__(self) -> None:
        self.paren_level = 0
        self.indent_level = [0]
        assert self.tab_len > 0

    def handle_NL(self, token: Token) -> Iterator[Token]:
        if self.paren_level > 0:
            return

        yield token

        indent_str = token.rsplit("\n", 1)[1]  ##

        indent = indent_str.count(" ") + indent_str.count("\t") * self.tab_len

        if indent > self.indent_level[-1]:
            self.indent_level.append(indent)
            yield Token.new_borrow_pos(self.INDENT_type, indent_str, token)
        else:
            while indent < self.indent_level[-1]:
                self.indent_level.pop()
                yield Token.new_borrow_pos(self.DEDENT_type, indent_str, token)

            if indent != self.indent_level[-1]:
                raise DedentError(
                    "Unexpected dedent to column %s. Expected dedent to %s"
                    % (indent, self.indent_level[-1])
                )

    def _process(self, stream):
        for token in stream:
            if token.type == self.NL_type:
                yield from self.handle_NL(token)
            else:
                yield token

            if token.type in self.OPEN_PAREN_types:
                self.paren_level += 1
            elif token.type in self.CLOSE_PAREN_types:
                self.paren_level -= 1
                assert self.paren_level >= 0

        while len(self.indent_level) > 1:
            self.indent_level.pop()
            yield Token(self.DEDENT_type, "")

        assert self.indent_level == [0], self.indent_level

    def process(self, stream):
        self.paren_level = 0
        self.indent_level = [0]
        return self._process(stream)

    ##

    @property
    def always_accept(self):
        return (self.NL_type,)

    @property
    @abstractmethod
    def NL_type(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def OPEN_PAREN_types(self) -> List[str]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def CLOSE_PAREN_types(self) -> List[str]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def INDENT_type(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def DEDENT_type(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def tab_len(self) -> int:
        raise NotImplementedError()


class PythonIndenter(Indenter):
    NL_type = "_NEWLINE"
    OPEN_PAREN_types = ["LPAR", "LSQB", "LBRACE"]
    CLOSE_PAREN_types = ["RPAR", "RSQB", "RBRACE"]
    INDENT_type = "_INDENT"
    DEDENT_type = "_DEDENT"
    tab_len = 8


import base64
import pickle
import zlib

DATA = {
    "parser": {
        "lexer_conf": {
            "terminals": [
                {"@": 0},
                {"@": 1},
                {"@": 2},
                {"@": 3},
                {"@": 4},
                {"@": 5},
                {"@": 6},
                {"@": 7},
            ],
            "ignore": ["WS"],
            "g_regex_flags": 0,
            "use_bytes": False,
            "lexer_type": "contextual",
            "__type__": "LexerConf",
        },
        "parser_conf": {
            "rules": [
                {"@": 8},
                {"@": 9},
                {"@": 10},
                {"@": 11},
                {"@": 12},
                {"@": 13},
                {"@": 14},
                {"@": 15},
                {"@": 16},
                {"@": 17},
                {"@": 18},
                {"@": 19},
                {"@": 20},
                {"@": 21},
                {"@": 22},
            ],
            "start": ["unit", "quantity"],
            "parser_type": "lalr",
            "__type__": "ParserConf",
        },
        "parser": {
            "tokens": {
                0: "__unit_sequence_plus_1",
                1: "_MULTIPLY",
                2: "SYMBOL",
                3: "$END",
                4: "_DIVIDE",
                5: "term",
                6: "__unit_sequence_plus_0",
                7: "unit_sequence",
                8: "exponent",
                9: "CARAT_EXPONENT",
                10: "SUPERSCRIPT_EXPONENT",
                11: "magnitude",
                12: "SIGNED_FLOAT",
                13: "quantity",
                14: "SIGNED_INT",
                15: "unit",
            },
            "states": {
                0: {
                    0: (0, 22),
                    1: (0, 6),
                    2: (1, {"@": 19}),
                    3: (1, {"@": 19}),
                    4: (1, {"@": 19}),
                },
                1: {3: (1, {"@": 22}), 4: (1, {"@": 22}), 1: (1, {"@": 22})},
                2: {2: (0, 10), 5: (0, 0), 6: (0, 7), 7: (0, 13)},
                3: {3: (1, {"@": 21}), 4: (1, {"@": 21}), 1: (1, {"@": 21})},
                4: {},
                5: {4: (0, 2), 3: (1, {"@": 12})},
                6: {5: (0, 3), 2: (0, 10)},
                7: {2: (0, 10), 5: (0, 11), 3: (1, {"@": 13}), 4: (1, {"@": 13})},
                8: {3: (1, {"@": 8})},
                9: {2: (0, 10), 5: (0, 1)},
                10: {
                    8: (0, 12),
                    9: (0, 17),
                    10: (0, 16),
                    2: (1, {"@": 16}),
                    3: (1, {"@": 16}),
                    1: (1, {"@": 16}),
                    4: (1, {"@": 16}),
                },
                11: {2: (1, {"@": 20}), 3: (1, {"@": 20}), 4: (1, {"@": 20})},
                12: {
                    2: (1, {"@": 15}),
                    3: (1, {"@": 15}),
                    1: (1, {"@": 15}),
                    4: (1, {"@": 15}),
                },
                13: {3: (1, {"@": 11})},
                14: {},
                15: {11: (0, 18), 12: (0, 21), 13: (0, 14), 14: (0, 20)},
                16: {
                    2: (1, {"@": 18}),
                    3: (1, {"@": 18}),
                    4: (1, {"@": 18}),
                    1: (1, {"@": 18}),
                },
                17: {
                    2: (1, {"@": 17}),
                    3: (1, {"@": 17}),
                    4: (1, {"@": 17}),
                    1: (1, {"@": 17}),
                },
                18: {6: (0, 7), 15: (0, 8), 5: (0, 0), 7: (0, 5), 2: (0, 10)},
                19: {6: (0, 7), 15: (0, 4), 5: (0, 0), 7: (0, 5), 2: (0, 10)},
                20: {2: (1, {"@": 9})},
                21: {2: (1, {"@": 10})},
                22: {1: (0, 9), 3: (1, {"@": 14}), 4: (1, {"@": 14})},
            },
            "start_states": {"unit": 19, "quantity": 15},
            "end_states": {"unit": 4, "quantity": 14},
        },
        "__type__": "ParsingFrontend",
    },
    "rules": [
        {"@": 8},
        {"@": 9},
        {"@": 10},
        {"@": 11},
        {"@": 12},
        {"@": 13},
        {"@": 14},
        {"@": 15},
        {"@": 16},
        {"@": 17},
        {"@": 18},
        {"@": 19},
        {"@": 20},
        {"@": 21},
        {"@": 22},
    ],
    "options": {
        "debug": False,
        "keep_all_tokens": False,
        "tree_class": None,
        "cache": False,
        "postlex": None,
        "parser": "lalr",
        "lexer": "contextual",
        "transformer": None,
        "start": ["unit", "quantity"],
        "priority": "normal",
        "ambiguity": "auto",
        "regex": False,
        "propagate_positions": False,
        "lexer_callbacks": {},
        "maybe_placeholders": False,
        "edit_terminals": None,
        "g_regex_flags": 0,
        "use_bytes": False,
        "import_paths": [],
        "source_path": None,
        "_plugins": {},
    },
    "__type__": "Lark",
}
MEMO = {
    0: {
        "name": "SIGNED_INT",
        "pattern": {
            "value": "(?:(?:\\+|\\-))?(?:[0-9])+",
            "flags": [],
            "_width": [1, 4294967295],
            "__type__": "PatternRE",
        },
        "priority": 0,
        "__type__": "TerminalDef",
    },
    1: {
        "name": "SIGNED_FLOAT",
        "pattern": {
            "value": "(?:(?:\\+|\\-))?(?:(?:[0-9])+(?:e|E)(?:(?:\\+|\\-))?(?:[0-9])+|(?:(?:[0-9])+\\.(?:(?:[0-9])+)?|\\.(?:[0-9])+)(?:(?:e|E)(?:(?:\\+|\\-))?(?:[0-9])+)?)",
            "flags": [],
            "_width": [2, 4294967295],
            "__type__": "PatternRE",
        },
        "priority": 0,
        "__type__": "TerminalDef",
    },
    2: {
        "name": "WS",
        "pattern": {
            "value": "(?:[ \t\x0c\r\n])+",
            "flags": [],
            "_width": [1, 4294967295],
            "__type__": "PatternRE",
        },
        "priority": 0,
        "__type__": "TerminalDef",
    },
    3: {
        "name": "_MULTIPLY",
        "pattern": {
            "value": "(?:⋅|\\*)",
            "flags": [],
            "_width": [1, 1],
            "__type__": "PatternRE",
        },
        "priority": 0,
        "__type__": "TerminalDef",
    },
    4: {
        "name": "_DIVIDE",
        "pattern": {"value": "/", "flags": [], "__type__": "PatternStr"},
        "priority": 0,
        "__type__": "TerminalDef",
    },
    5: {
        "name": "CARAT_EXPONENT",
        "pattern": {
            "value": "\\^(?:(?:\\+|\\-))?(?:[0-9])+",
            "flags": [],
            "_width": [2, 4294967295],
            "__type__": "PatternRE",
        },
        "priority": 0,
        "__type__": "TerminalDef",
    },
    6: {
        "name": "SUPERSCRIPT_EXPONENT",
        "pattern": {
            "value": "(?:⁻)?(?:(?:⁰|¹|²|³|⁴|⁵|⁶|⁷|⁸|⁹))+",
            "flags": [],
            "_width": [1, 4294967295],
            "__type__": "PatternRE",
        },
        "priority": 0,
        "__type__": "TerminalDef",
    },
    7: {
        "name": "SYMBOL",
        "pattern": {
            "value": "(?:(?:(?:\\.|°|\\-|\\(|\\))|(?:[A-Z]|Å)|[a-z]|[ₐ-ₜ]|[Α-ω]|1|☉))+",
            "flags": [],
            "_width": [1, 4294967295],
            "__type__": "PatternRE",
        },
        "priority": 0,
        "__type__": "TerminalDef",
    },
    8: {
        "origin": {"name": Token("RULE", "quantity"), "__type__": "NonTerminal"},
        "expansion": [
            {"name": "magnitude", "__type__": "NonTerminal"},
            {"name": "unit", "__type__": "NonTerminal"},
        ],
        "order": 0,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    9: {
        "origin": {"name": Token("RULE", "magnitude"), "__type__": "NonTerminal"},
        "expansion": [
            {"name": "SIGNED_INT", "filter_out": False, "__type__": "Terminal"}
        ],
        "order": 0,
        "alias": "int",
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    10: {
        "origin": {"name": Token("RULE", "magnitude"), "__type__": "NonTerminal"},
        "expansion": [
            {"name": "SIGNED_FLOAT", "filter_out": False, "__type__": "Terminal"}
        ],
        "order": 1,
        "alias": "float",
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    11: {
        "origin": {"name": Token("RULE", "unit"), "__type__": "NonTerminal"},
        "expansion": [
            {"name": "unit_sequence", "__type__": "NonTerminal"},
            {"name": "_DIVIDE", "filter_out": True, "__type__": "Terminal"},
            {"name": "unit_sequence", "__type__": "NonTerminal"},
        ],
        "order": 0,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    12: {
        "origin": {"name": Token("RULE", "unit"), "__type__": "NonTerminal"},
        "expansion": [{"name": "unit_sequence", "__type__": "NonTerminal"}],
        "order": 1,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    13: {
        "origin": {"name": Token("RULE", "unit_sequence"), "__type__": "NonTerminal"},
        "expansion": [{"name": "__unit_sequence_plus_0", "__type__": "NonTerminal"}],
        "order": 0,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    14: {
        "origin": {"name": Token("RULE", "unit_sequence"), "__type__": "NonTerminal"},
        "expansion": [
            {"name": "term", "__type__": "NonTerminal"},
            {"name": "__unit_sequence_plus_1", "__type__": "NonTerminal"},
        ],
        "order": 1,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    15: {
        "origin": {"name": Token("RULE", "term"), "__type__": "NonTerminal"},
        "expansion": [
            {"name": "SYMBOL", "filter_out": False, "__type__": "Terminal"},
            {"name": "exponent", "__type__": "NonTerminal"},
        ],
        "order": 0,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    16: {
        "origin": {"name": Token("RULE", "term"), "__type__": "NonTerminal"},
        "expansion": [{"name": "SYMBOL", "filter_out": False, "__type__": "Terminal"}],
        "order": 1,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    17: {
        "origin": {"name": Token("RULE", "exponent"), "__type__": "NonTerminal"},
        "expansion": [
            {"name": "CARAT_EXPONENT", "filter_out": False, "__type__": "Terminal"}
        ],
        "order": 0,
        "alias": "carat_exponent",
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    18: {
        "origin": {"name": Token("RULE", "exponent"), "__type__": "NonTerminal"},
        "expansion": [
            {
                "name": "SUPERSCRIPT_EXPONENT",
                "filter_out": False,
                "__type__": "Terminal",
            }
        ],
        "order": 1,
        "alias": "superscript_exponent",
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    19: {
        "origin": {"name": "__unit_sequence_plus_0", "__type__": "NonTerminal"},
        "expansion": [{"name": "term", "__type__": "NonTerminal"}],
        "order": 0,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    20: {
        "origin": {"name": "__unit_sequence_plus_0", "__type__": "NonTerminal"},
        "expansion": [
            {"name": "__unit_sequence_plus_0", "__type__": "NonTerminal"},
            {"name": "term", "__type__": "NonTerminal"},
        ],
        "order": 1,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    21: {
        "origin": {"name": "__unit_sequence_plus_1", "__type__": "NonTerminal"},
        "expansion": [
            {"name": "_MULTIPLY", "filter_out": True, "__type__": "Terminal"},
            {"name": "term", "__type__": "NonTerminal"},
        ],
        "order": 0,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
    22: {
        "origin": {"name": "__unit_sequence_plus_1", "__type__": "NonTerminal"},
        "expansion": [
            {"name": "__unit_sequence_plus_1", "__type__": "NonTerminal"},
            {"name": "_MULTIPLY", "filter_out": True, "__type__": "Terminal"},
            {"name": "term", "__type__": "NonTerminal"},
        ],
        "order": 1,
        "alias": None,
        "options": {
            "keep_all_tokens": False,
            "expand1": False,
            "priority": None,
            "template_source": None,
            "empty_indices": (),
            "__type__": "RuleOptions",
        },
        "__type__": "Rule",
    },
}
Shift = 0
Reduce = 1


def Parser(**kwargs):
    return Lark._load_from_dict(DATA, MEMO, **kwargs)



from measured import Acceleration, Crackle, Jerk, Length, Pop, Snap, Speed
from measured.si import Meter, Second

# https://en.wikipedia.org/wiki/Fourth,_fifth,_and_sixth_derivatives_of_position


def test_derivatives_of_position() -> None:
    length = 128 * Meter
    assert length.unit.dimension is Length

    speed = length / (2 * Second)
    assert speed.unit.dimension is Speed
    assert speed == 64 * Meter / Second

    acceleration = speed / (2 * Second)
    assert acceleration.unit.dimension is Acceleration
    assert acceleration == 32 * Meter / Second**2

    jerk = acceleration / (2 * Second)
    assert jerk.unit.dimension is Jerk
    assert jerk == 16 * Meter / Second**3

    snap = jerk / (2 * Second)
    assert snap.unit.dimension is Snap
    assert snap == 8 * Meter / Second**4

    crackle = snap / (2 * Second)
    assert crackle.unit.dimension is Crackle
    assert crackle == 4 * Meter / Second**5

    pop = crackle / (2 * Second)
    assert pop.unit.dimension is Pop
    assert pop == 2 * Meter / Second**6



import pytest
from hypothesis import example, given

from measured import IdentityPrefix, One, Prefix, Quantity, approximately
from measured.hypothesis import prefixes
from measured.iec import Bit, Kibi, Mebi
from measured.si import Ampere, Deci, Kilo, Mega, Meter, Micro, Milli, Ohm, Second, Volt


@pytest.fixture(scope="module")
def identity() -> Prefix:
    return IdentityPrefix


@given(a=prefixes())
def test_abelian_identity(identity: Prefix, a: Prefix) -> None:
    assert identity * a == a


@given(a=prefixes())
@example(a=IdentityPrefix)
def test_abelian_inverse(identity: Prefix, a: Prefix) -> None:
    inverse = a**-1
    assert inverse * a == a * inverse
    assert inverse * a == identity
    assert identity / a == inverse


@given(prefix=prefixes())
def test_repr(prefix: Prefix) -> None:
    assert repr(prefix) == f"Prefix(base={prefix.base!r}, exponent={prefix.exponent!r})"


@given(prefix=prefixes())
def test_repr_roundtrips(prefix: Prefix) -> None:
    assert eval(repr(prefix)) is prefix


def test_prefixes_scale_quantities_up() -> None:
    length = 3 * (Kilo * Meter)
    assert length == 3000 * Meter


def test_prefixes_scale_quantities_down() -> None:
    length = 3 * (Micro * Meter)
    assert length == 0.000003 * Meter


def test_equal_prefixes_cancel() -> None:
    speed = (15 * (Kilo * Meter)) / (3 * (Kilo * Second))
    assert speed == (5 * (Kilo * Meter)) / (Kilo * Second)
    assert speed == (5000 * Meter) / (1000 * Second)
    assert speed == 5 * Meter / Second


def test_different_prefixes_cancel() -> None:
    speed = (15 * (Mega * Meter)) / (3 * (Kilo * Second))
    assert speed == (15 * (Kilo * Meter)) / (3 * Second)
    assert speed == (5 * (Kilo * Meter)) / Second
    assert speed == 5000 * Meter / Second


def test_can_compare_quantities_with_different_prefixes() -> None:
    assert 1.0 * (Kibi * Bit) == 1.024 * (Kilo * Bit)


def test_can_multiply_prefixes_with_same_base() -> None:
    assert 5 * ((Kilo * Mega) * Meter) == 5000000000 * Meter


def test_can_multiply_prefixes_with_different_bases() -> None:
    iec_first = 5 * (Kibi * Mega) * Meter
    si_first = 5 * (Mega * Kibi) * Meter
    assert iec_first == approximately(si_first)
    assert iec_first == approximately(5120000000 * Meter)
    assert si_first == approximately(5120000000 * Meter)


def test_can_divide_prefixes_with_same_base() -> None:
    assert 5 * (Mega / Kilo) * Meter == 5000 * Meter


def test_can_divide_prefixes_with_different_bases() -> None:
    assert 5 * (Mebi / Kilo) * Meter == approximately(5242.88 * Meter)
    assert 5 * (Mega / Kibi) * Meter == approximately(4882.8125 * Meter)


def test_multiplying_produces_number_quantities() -> None:
    assert 5 * Kilo == 5000 * One
    assert 5 * Milli == 0.005 * One


def test_associative_with_multiplication() -> None:
    assert (5 * Kilo) * Meter == 5 * (Kilo * Meter)


def test_associative_in_denominator() -> None:
    assert 10000 * Meter / (5 * Kilo * Second) == 2 * Meter / Second


def test_multiplying_by_random_things() -> None:
    with pytest.raises(TypeError):
        "hello" * Kilo  # type: ignore

    with pytest.raises(TypeError):
        Kilo * "hello"  # type: ignore


def test_dividing_by_random_things() -> None:
    with pytest.raises(TypeError):
        "hello" / Kilo  # type: ignore

    with pytest.raises(TypeError):
        Kilo / "hello"  # type: ignore


def test_prefixed_complex_units() -> None:
    voltage = 5 * Kilo * Volt
    current = 2 * Milli * Ampere
    resistance = voltage / current
    assert resistance == 2.5 * Mega * Ohm
    assert resistance == 2500000 * Ohm


@pytest.mark.parametrize(
    "root, prefix_squared, squared",
    [
        (
            (5 * Kilo * Meter),
            25 * ((Kilo * Meter) ** 2),
            25000000 * Meter**2,
        ),
        (
            (100 * Kilo * Meter) / (10 * Second),
            (10000 * ((Kilo * Meter) ** 2)) / (100 * Second**2),
            (10000000000 * Meter**2) / (100 * Second**2),
        ),
    ],
)
def test_prefixed_unit_exponentation(
    root: Quantity, prefix_squared: Quantity, squared: Quantity
) -> None:
    assert root**2 == prefix_squared
    assert root**2 == squared


def test_roots() -> None:
    assert Mega.root(0) == IdentityPrefix
    assert Mega.root(2) == Kilo
    assert Mebi.root(2) == Kibi


def test_only_integer_roots() -> None:
    with pytest.raises(TypeError):
        Deci.root(0.5)  # type: ignore


def test_whole_power_roots_only() -> None:
    with pytest.raises(ValueError):
        Deci.root(3)


def test_prefixes_invert_properly() -> None:
    assert Mega * (Meter**-1) == (Micro * Meter) ** -1
    assert (Micro * Meter) ** -1 == Mega * (Meter**-1)



from measured import approximately
from measured.apocrypha import Smoot
from measured.si import Meter
from measured.us import Foot


def test_smoot() -> None:
    assert 364.4 * Smoot == approximately(2035 * Foot, within=3e-4)
    assert 364.4 * Smoot == approximately(620.1 * Meter, within=3e-4)



"""
Units and non-standardized conventions for representing `Energy` and `Power`.

Attributes: Units of Energy

    Erg (Unit):

    Calorie (Unit):

    BritishThermalUnit (Unit):

    TonOfRefrigeration (Unit):

    TonneOfTNT (Unit):


Attributes: Units of Power

    Horsepower (Unit):

    Donkeypower (Unit):

    MetricHorsepower (Unit):

    ElectricalHorsepower (Unit):

    BoilerHorsepower (Unit):


Attributes: Other units

    Langley (Unit): A unit of heat transmission (`Energy/Area`), often used to express
        insolation (the amount of solar energy reaching the surface of the Earth)

"""

from measured import Area, Energy, Power
from measured.si import Hour, Joule, Kilo, Meter, Second, Watt
from measured.us import Foot, PoundForce

# https://en.wikipedia.org/wiki/Erg
Erg = Energy.unit(name="erg", symbol="erg")
Erg.equals(1e-7 * Joule)

# https://en.wikipedia.org/wiki/Calorie
Calorie = Energy.unit(name="calorie", symbol="cal")
Calorie.equals(4.184 * Joule)

# https://en.wikipedia.org/wiki/British_thermal_unit#Definitions
BritishThermalUnit = Energy.unit(name="British thermal unit", symbol="BTU")
BritishThermalUnit.equals(1054.350264488889 * Joule)

# https://en.wikipedia.org/wiki/Ton_of_refrigeration
TonOfRefrigeration = Power.unit(name="ton of refrigeration", symbol="TR")
TonOfRefrigeration.equals(12000 * BritishThermalUnit / Hour)
TonOfRefrigeration.equals(3.51685 * Kilo * Watt)

# https://en.wikipedia.org/wiki/TNT_equivalent
TonneOfTNT = Energy.unit(name="ton of TNT", symbol="tₜₙₜ")
TonneOfTNT.equals(4.184e9 * Joule)
TonneOfTNT.equals(1e9 * Calorie)


# https://en.wikipedia.org/wiki/Horsepower#Definitions
Horsepower = Power.unit("horsepower", "hp")
Horsepower.equals(550 * Foot * PoundForce / Second)

Donkeypower = Power.unit("donkeypower", "donkeypower")
Donkeypower.equals(1 / 3 * Horsepower)

MetricHorsepower = Power.unit("metric horsepower", "hp(M)")
MetricHorsepower.equals(735.49875 * Watt)

ElectricalHorsepower = Power.unit("electrical horsepower", "hp(E)")
ElectricalHorsepower.equals(746 * Watt)

BoilerHorsepower = Power.unit("boiler horsepower", "hp(S)")
BoilerHorsepower.equals(33475 * BritishThermalUnit / Hour)


# https://en.wikipedia.org/wiki/Langley_(unit)
Langley = (Energy / Area).unit("langley", "Ly")
Langley.equals(41840 * Joule / Meter**2)



import pytest

from measured import Charge, Length, Mass, Quantity, Time, approximately
from measured.natural import StoneyCharge, StoneyLength, StoneyMass, StoneyTime
from measured.physics import G, c, e, kₑ, α, ℏ
from measured.si import Coulomb, Kilogram, Meter, Second


@pytest.mark.parametrize(
    "fundamental_constant, unity",
    [
        (c, 1 * (StoneyLength / StoneyTime)),
        (G, 1 * (StoneyLength**3 / (StoneyMass * StoneyTime**2))),
        (e, 1 * StoneyCharge),
        (
            kₑ,
            1
            * (
                (StoneyLength**3 * StoneyMass) / (StoneyTime**2 * StoneyCharge**2)
            ),
        ),
        (ℏ, α**-1 * ((StoneyLength**2 * StoneyMass) / StoneyTime)),
    ],
)
def test_unity(fundamental_constant: Quantity, unity: Quantity) -> None:
    assert fundamental_constant == approximately(unity)
    assert unity == approximately(fundamental_constant)


def test_length() -> None:
    assert StoneyLength.dimension is Length
    assert 1 * StoneyLength == approximately(1.3807e-36 * Meter, within=2e-5)


def test_mass() -> None:
    assert StoneyMass.dimension is Mass
    assert 1 * StoneyMass == approximately(1.8592e-9 * Kilogram, within=5e-6)


def test_time() -> None:
    assert StoneyTime.dimension is Time
    assert 1 * StoneyTime == approximately(4.6054e-45 * Second, within=2e-5)


def test_charge() -> None:
    assert StoneyCharge.dimension is Charge
    assert 1 * StoneyCharge == approximately(1.6022e-19 * Coulomb, within=2e-5)



import pytest

from measured.pytest import MeasuredPlugin
from measured.si import Meter, Second


@pytest.fixture
def plugin() -> MeasuredPlugin:
    return MeasuredPlugin()


def test_explaining_assertions(plugin: MeasuredPlugin) -> None:
    explanation = plugin.pytest_assertrepr_compare(
        op="==",
        left=1 * Meter,
        right=1 * Second,
    )
    assert isinstance(explanation, list)
    assert all(isinstance(line, str) for line in explanation)


def test_only_explains_measured_things(plugin: MeasuredPlugin) -> None:
    explanation = plugin.pytest_assertrepr_compare(op="==", left=1, right=2)
    assert explanation is None



import pytest

from measured import (
    Acceleration,
    Area,
    Capacitance,
    Charge,
    Dimension,
    Energy,
    Force,
    Length,
    Mass,
    Number,
    One,
    Quantity,
    Speed,
    Temperature,
    Time,
    Volume,
    approximately,
    physics,
)
from measured.si import (
    Coulomb,
    Dalton,
    ElectronVolt,
    Farad,
    Joule,
    Kelvin,
    Kilogram,
    Mega,
    Meter,
    Second,
    Volt,
)
from measured.us import Foot


@pytest.mark.parametrize(
    "constant, dimension",
    [
        (physics.c, Speed),
        (physics.G, Volume / (Mass * Time**2)),
        (physics.h, Energy * Time),
        (physics.ℏ, Energy * Time),
        (physics.k, Energy / Temperature),
        (physics.e, Charge),
        (physics.ε0, Capacitance / Length),
        (physics.mₑ, Mass),
        (physics.kₑ, ((Force * Area) / Charge**2)),
        (physics.α, Number),
        (physics.gₙ, Acceleration),
    ],
)
def test_constants_have_expected_dimension(
    constant: Quantity, dimension: Dimension
) -> None:
    assert constant.unit.dimension is dimension


# https://en.wikipedia.org/wiki/Physical_constant#Table_of_physical_constants
# While these are redundant (quite a few are _defined_ in Si units), this test is
# intended to give flexibility about redefining constants in terms of others.  This test
# confirms that no matter what, we can always depend on their SI conversions.
@pytest.mark.parametrize(
    "constant, expected",
    [
        (physics.c, 299792458 * Meter / Second),
        (physics.G, 6.67430e-11 * Meter**3 / (Kilogram * Second**2)),
        (physics.h, 6.62607015e-34 * Joule * Second),
        (physics.ℏ, 1.0545718176461565e-34 * Joule * Second),
        (physics.k, 1.380649e-23 * Joule / Kelvin),
        (physics.e, 1.602176634e-19 * Coulomb),
        (physics.ε0, 8.8541878128e-12 * Farad / Meter),
        (physics.mₑ, 9.1093837015e-31 * Kilogram),
        (physics.mₑ, 0.0005485799090649074 * Dalton),
        (physics.e, 1.602176634e-19 * Coulomb),
        (
            physics.kₑ,
            8.9875517923e9 * (Kilogram * Meter**3) / (Second**2 * Coulomb**2),
        ),
        (physics.α, 1 / 137.035999206 * One),
        (physics.gₙ, 9.80665 * Meter / Second**2),
        (physics.gₙ, 32.174049 * Foot / Second**2),
    ],
)
def test_constants_have_expected_si_values(
    constant: Quantity, expected: Quantity
) -> None:
    assert constant == approximately(expected)


def test_electronvolt() -> None:
    c = physics.c
    e = physics.e

    assert 1 * ElectronVolt == e * Volt
    assert 1 * ElectronVolt / c**2 == approximately(1.78266192e-36 * Kilogram)
    assert 1 * (Mega * ElectronVolt) / c**2 == approximately(0.0010735441 * Dalton)



from measured import Area, approximately
from measured.metric import Barn, Outhouse, Shake, Shed
from measured.si import Meter, Micro, Nano, Second, Yocto


def test_barn() -> None:
    assert Barn.dimension is Area
    assert Barn.name == "barn"
    assert Barn.symbol == "barn"  # we already have a 'b' symbol at measured.iec.Bit
    assert 1 * Barn == approximately(1e-28 * Meter**2)


def test_outhouse() -> None:
    assert Outhouse.dimension is Area
    assert Outhouse.name == "outhouse"
    assert Outhouse.symbol == "outhouse"
    assert 1 * Outhouse == approximately(1 * Micro * Barn)
    assert 1 * Outhouse == approximately(1e-34 * Meter**2)


def test_shed() -> None:
    assert Shed.dimension is Area
    assert Shed.name == "shed"
    assert Shed.symbol == "shed"
    assert 1 * Shed == approximately(1 * Yocto * Barn)
    assert 1 * Shed == approximately(1e-52 * Meter**2)


def test_shake() -> None:
    assert 1 * Shake == 10 * Nano * Second



import sys
import textwrap
from argparse import ArgumentParser, RawTextHelpFormatter
from typing import Generator, Iterable, Optional, Set, Tuple

from measured import Quantity, Unit, conversions, systems  # noqa: F401

parser = ArgumentParser(
    description="Unit conversions with measured",
    epilog=textwrap.dedent(
        """
        Quantities are expressed as a number followed by any number of terms multiplied,
        divided, or raised to powers.  For example:

        $ measured 5 mile
        $ measured 3 meter^2
        $ measured 4 meter/second
    """,
    ),
    formatter_class=RawTextHelpFormatter,
)
parser.add_argument(
    "quantity", metavar="term", nargs="*", help="The quantity to print conversions for"
)
parser.add_argument("--list", action="store_true", help="List all available units")


def all_equivalents(
    quantity: Quantity,
    depth: int = 0,
    visited: Optional[Set[Unit]] = None,
) -> Generator[Quantity, None, None]:
    unit = quantity.unit

    if visited is None:
        visited = {unit}
    else:
        visited.add(unit)

    for other, ratio in conversions._ratios[unit].items():
        if other in visited:
            continue

        offset = conversions._offsets[unit].get(other, 0)

        magnitude = (quantity.magnitude * ratio) + offset
        next_quantity = Quantity(magnitude, other)

        yield next_quantity
        yield from all_equivalents(next_quantity, depth + 1, visited=visited)


def dot_aligned(
    quantities: Iterable[Quantity],
) -> Generator[Tuple[str, Unit], None, None]:
    try:
        magnitudes, units = zip(*((q.magnitude, q.unit) for q in quantities))
    except ValueError:
        return

    magnitude_strings = [str(m) for m in magnitudes]
    dot_indices = [s.find(".") for s in magnitude_strings]
    largest_left = max(dot_indices)
    right_padding = max(len(m) for m in magnitude_strings) + largest_left
    for (magnitude_string, dot_index), unit in zip(
        zip(magnitude_strings, dot_indices), units
    ):
        left_padding = largest_left - dot_index
        yield format(" " * left_padding + magnitude_string, f"<{right_padding}"), unit


def print_unit_list() -> None:
    for symbol, unit in Unit._by_symbol.items():
        print(f"{symbol} ({unit.name}, {unit.dimension.name or unit.dimension.symbol})")


def print_quantity(input_string: str) -> None:
    try:
        quantity = Quantity.parse(input_string)
    except Exception as e:
        print(f"Error parsing quantity from {input_string!r}:\n{str(e)}")
        sys.exit(1)

    quantity = quantity.unprefixed()
    unit = quantity.unit
    dimension = unit.dimension

    print("Magnitude:", quantity.magnitude)
    print("Unit:", unit.name or f"{unit:/}")
    print("Dimension:", dimension.name or f"{dimension}")

    print("Equivalent to:")
    equivalents = sorted(
        all_equivalents(quantity), key=lambda q: abs(q.magnitude), reverse=True
    )
    for magnitude_string, unit in dot_aligned(equivalents):
        print(f"{magnitude_string} {unit.name or unit.__format__('/')}")


def main() -> None:
    arguments = parser.parse_args()

    if arguments.list:
        print_unit_list()
        return

    input_string = " ".join(arguments.quantity)
    if input_string:
        print_quantity(input_string)
        return

    parser.print_help()



# https://en.wikipedia.org/wiki/FFF_system

from measured import Mass, Time, us
from measured.si import Day

Furlong = us.Furlong

Firkin = Mass.unit("firkin", "fir")
Firkin.equals(90 * us.Pound)

Fortnight = Time.unit("fortnight", "ftn")
Fortnight.equals(14 * Day)



import pytest

from measured import Quantity, approximately
from measured.si import Kelvin, Kilo, Milli
from measured.us import Fahrenheit, Rankine


@pytest.mark.parametrize(
    "kelvin, rankine",
    [
        # Absolute Zero
        (0 * Kelvin, 0 * Rankine),
        # Combined with prefixes > 1
        (10 * Kilo * Kelvin, 18 * Kilo * Rankine),
        (10 * Kilo * Kelvin, 18000 * Rankine),
        (10000 * Kelvin, 18 * Kilo * Rankine),
    ],
)
def test_kelvin_and_rankine(kelvin: Quantity, rankine: Quantity) -> None:
    assert kelvin == rankine
    assert rankine == kelvin


@pytest.mark.parametrize(
    "kelvin, rankine",
    [
        # Freezing point of brine
        (255.37 * Kelvin, 459.665999 * Rankine),
        # Freezing point of water
        (273.15 * Kelvin, 491.67 * Rankine),
        # Boiling point of water
        (373.1339 * Kelvin, 671.64102 * Rankine),
        # Combined with prefixes < 1
        (10 * Milli * Kelvin, 18 * Milli * Rankine),
        (10 * Milli * Kelvin, 0.018 * Rankine),
        (0.01 * Kelvin, 18 * Milli * Rankine),
    ],
)
def test_kelvin_approximates_rankine(kelvin: Quantity, rankine: Quantity) -> None:
    assert kelvin == approximately(rankine)
    assert rankine == approximately(kelvin)


@pytest.mark.parametrize(
    "rankine, fahrenheit",
    [
        (0 * Rankine, -459.67 * Fahrenheit),
        (459.67 * Rankine, 0 * Fahrenheit),
    ],
)
def test_rankine_equals_fahrenheit(rankine: Quantity, fahrenheit: Quantity) -> None:
    assert rankine == fahrenheit
    assert fahrenheit == rankine


@pytest.mark.parametrize(
    "rankine, fahrenheit",
    [
        (459.68 * Rankine, 10 * Milli * Fahrenheit),
        (559.67 * Rankine, 100 * Fahrenheit),
        (559670.0 * Milli * Rankine, 100 * Fahrenheit),
    ],
)
def test_rankine_approximates_fahrenheit(
    rankine: Quantity, fahrenheit: Quantity
) -> None:
    assert rankine == approximately(fahrenheit)
    assert fahrenheit == approximately(rankine)



import pytest

from measured import Quantity, Volume, approximately
from measured.si import Centi, Kilo, Liter, Meter, Milli
from measured.us import (
    Acre,
    AcreFoot,
    BoardFoot,
    Cord,
    Cup,
    FluidOunce,
    Foot,
    Gallon,
    Hogshead,
    Inch,
    Mile,
    Rick,
    Rod,
    Tablespoon,
    Teaspoon,
)


def test_common_cooking_measures() -> None:
    assert 1 * Teaspoon == 4.92892159375 * Milli * Liter

    assert 1 * Tablespoon == 14.78676478125 * Milli * Liter
    assert 1 * Tablespoon == 3 * Teaspoon

    assert 1 * Cup == 236.5882365 * Milli * Liter
    assert 1 * Cup == 8 * FluidOunce


@pytest.mark.parametrize(
    "left, right",
    [
        (1 * Foot**3, 1728 * Inch**3),
        (1 * Mile / Foot**3, 1 / 1728 * Mile / Inch**3),
        (10 * Gallon / Mile, 0.023521459 * Liter / Meter),
        (10 * Gallon / Mile, 23.521459 * Liter / (Kilo * Meter)),
        (10 * Mile / Gallon, 4251.4370 * Meter / Liter),
        (10 * Mile / Gallon, 4.2514370 * (Kilo * Meter) / Liter),
    ],
)
def test_volume_conversions(left: Quantity, right: Quantity) -> None:
    assert left == approximately(right)
    assert right == approximately(left)


def test_abe() -> None:
    abes_car = (40 * Rod) / (1 * Hogshead)

    assert (40 * Rod) == approximately(0.125 * Mile)
    assert 1 * Hogshead == 63 * Gallon

    assert abes_car == approximately((0.125 / 63) * Mile / Gallon, 0)
    assert 0.00198 * Mile / Gallon <= abes_car <= 0.00199 * Mile / Gallon


def test_boardfoot() -> None:
    assert BoardFoot.dimension is Volume
    assert BoardFoot.name == "boardfoot"
    assert BoardFoot.symbol == "FBM"
    assert 1 * BoardFoot == 12 * 12 * 1 * Inch**3
    assert 1 * BoardFoot == approximately(2359.737 * (Centi * Meter) ** 3)
    assert 1 * BoardFoot == approximately(0.002359737 * Meter**3)
    assert 1 * BoardFoot == approximately(2.359737 * Liter)


def test_cord_and_rick() -> None:
    assert 1 * Cord == 3 * Rick
    assert 1 * Cord == approximately(3.623 * Meter**3, within=5e-4)


def test_acrefoot() -> None:
    assert AcreFoot.dimension is Volume
    assert AcreFoot.name == "acre-foot"
    assert AcreFoot.symbol == "acre-ft."
    assert 1 * AcreFoot == 1 * Acre * Foot
    assert 1 * AcreFoot == approximately(325851 * Gallon, within=2e-6)
    assert 1 * AcreFoot == approximately(1233 * Meter**3, within=4e-4)



from measured import Acceleration, Energy, Force, Mass, Power, Pressure
from measured.si import Gram, Joule, Kilo, Meter, Milli, Newton, Pascal, Second, Watt


def test_newton() -> None:
    assert Newton.name == "newton"
    assert Newton.symbol == "N"
    assert Newton.dimension is Force


def test_pascal() -> None:
    assert Pascal.name == "pascal"
    assert Pascal.symbol == "Pa"
    assert Pascal.dimension is Pressure


def test_joule() -> None:
    assert Joule.name == "joule"
    assert Joule.symbol == "J"
    assert Joule.dimension is Energy


def test_watt() -> None:
    assert Watt.name == "watt"
    assert Watt.symbol == "W"
    assert Watt.dimension is Power


def test_newtons_second_law_gram_meter() -> None:
    mass = 5 * Gram
    assert mass.unit.dimension is Mass

    acceleration = 3 * Meter / Second**2
    assert acceleration.unit.dimension is Acceleration

    force = mass * acceleration
    assert force.unit.dimension is Force

    assert force == 15 * Gram * Meter / Second**2
    assert force == 0.015 * Newton


def test_newtons_second_law_kilogram_millimeter() -> None:
    mass = 5 * (Kilo * Gram)
    assert mass.unit.dimension is Mass

    acceleration = 3 * (Milli * Meter) / Second**2
    assert acceleration.unit.dimension is Acceleration

    force = mass * acceleration
    assert force.unit.dimension is Force

    assert force == 15 * Gram * Meter / Second**2
    assert force == 0.015 * Newton


def test_energy_is_force_through_distance() -> None:
    force = 10 * Newton
    distance = 5 * Meter
    energy = force * distance
    assert energy.unit == Joule
    assert energy.unit.dimension is Energy


def test_power_is_energy_per_time() -> None:
    energy = 10 * Joule
    power = energy / (2 * Second)
    assert power.unit == Watt
    assert power.unit.dimension is Power



from typing import Dict

from measured import Level, Octave, Prefix
from measured.iso import A440

Semitone = (Prefix(12, -1) * Octave).alias(name="semitone", symbol="semitone")

ConcertPitch = Semitone[A440]

# https://en.wikipedia.org/wiki/Scientific_pitch_notation

SPN_OCTAVES = range(-1, 11)
MIDDLE_A_OCTAVE = 4


def _octaves(note_offset: int) -> Dict[int, Level]:
    return {
        o: (((-MIDDLE_A_OCTAVE + o) * 12) + note_offset) * ConcertPitch
        for o in SPN_OCTAVES
    }


A_flat = _octaves(-1)
A = _octaves(0)
A_sharp = _octaves(+1)
B_flat = A_sharp
B = _octaves(+2)
C = _octaves(-9)
C_sharp = _octaves(-8)
D_flat = C_sharp
D = _octaves(-7)
D_sharp = _octaves(-6)
E_flat = D_sharp
E = _octaves(-5)
F = _octaves(-4)
F_sharp = _octaves(-3)
G_flat = F_sharp
G = _octaves(-2)
G_sharp = A_flat

MIDI = [(C[-1].magnitude + i) * ConcertPitch for i in range(128)]



from measured.computing import Furman, Nibble
from measured.geometry import π
from measured.iec import Bit, Byte
from measured.si import Arcsecond, Degree, Radian


def test_furman() -> None:
    assert 65536 * Furman == 2 * π * Radian
    assert 65536 * Furman == 360 * Degree
    assert 19.77 * Arcsecond < 1 * Furman < 20 * Arcsecond


def test_nibblee() -> None:
    assert 1 * Nibble == 4 * Bit
    assert 2 * Nibble == 1 * Byte
